import os, json, shutil, glob
from requests import get as requests_get
from pathlib import Path as pathlib_Path
from collections import namedtuple

from .colours import cl

if os.name != 'posix':
    try:
        from psutil import Process
        use_psutil = True
    except ImportError:  # psutil is nice to have but not necessary
        use_psutil = False


def rm_ext(fp):
    return '.'.join(fp.split('.')[:-1])

def replace_ext(fp, new_ext):
    return '.'.join(fp.split('.')[:-1])+'.'+new_ext

def parent_dir(fp):
    return '/'.join(fp.split('/')[:-1])

def lsfiles(root_dir, **kwargs):
    blacklist_ext=[]
    whitelist_ext=[]
    whitelist_all=True
    recursive=False
    ext_case_sensitive=False
    full_paths=True
    relative_paths=False
    if 'blacklist_ext' in kwargs:
        blacklist_ext=kwargs['blacklist_ext']
        if type(blacklist_ext)==type(''):
            blacklist_ext=blacklist_ext.split(' ')
    if 'whitelist_ext' in kwargs:
        whitelist_ext=kwargs['whitelist_ext']
        if type(whitelist_ext)==type(''):
            whitelist_ext=whitelist_ext.split(' ')
    if 'whitelist_all' in kwargs:
        whitelist_all=kwargs['whitelist_all']
    if 'recursive' in kwargs:
        recursive=kwargs['recursive']
    if 'full_paths' in kwargs:
        full_paths=kwargs['full_paths']
    if 'relative_paths' in kwargs:
        relative_paths=kwargs['relative_paths']
    if 'paths' in kwargs:
        if kwargs['paths']=='relative':
            relative_paths=True
        elif kwargs['paths']=='full':
            full_paths=True
    
    if (full_paths)&(relative_paths):
        raise Exception('full_paths and relative_paths flags are mutually exclusive')
    
    
    fp_ls=[]
    if 'glob' in kwargs:
        # lsfiles('/tmp', glob='foo*.json')
        fp_ls=glob.glob(os.path.join(root_dir, kwargs['glob']))
    elif recursive:
        for path, dirs, files in os.walk(root_dir):
            filepaths=[]
            for fname in files:
                ext=fname.split('.')[-1].lower()
                if (ext in whitelist_ext)|((ext not in blacklist_ext)&(whitelist_all)):
                    if full_paths:
                        filepaths.append(os.path.join(path, fname))
                    elif relative_paths:
                        dummy=os.path.join(path, fname)
                        filepaths.append(dummy.split(root_dir)[-1])
                    else:
                        filepaths.append(fname)
            fp_ls=fp_ls+filepaths
    else:
        fp_ls=os.listdir(root_dir)
        if full_paths:
            fp_ls=[os.path.join(root_dir, fname) for fname in fp_ls]
        if not whitelist_all:
            fp_ls=[fp for fp in fp_ls if fp.split('.')[-1].lower() in whitelist_ext]
        fp_ls=[fp for fp in fp_ls if fp.split('.')[-1].lower() not in blacklist_ext]
    for i in range(len(fp_ls)):
        while '/./' in fp_ls[i]:
            fp_ls[i]=fp_ls[i].replace('/./', '/')
    
    if 'sort' in kwargs:
        if kwargs['sort'] in ['modified', 'mtime']:
            fp_ls.sort(key=lambda x: os.path.getmtime(x))
    
    return fp_ls

def lsdir(path, **kwargs):
    # Look into using os.scandir instead of os.lsdir, better performance in many cases
    # Merge with lsfiles?
    paths=os.listdir(path)
    if 'items' in kwargs:
        item_n=int(kwargs['items'][1:])
        if kwargs['items'][0]=='>':
            dirls=[pth for pth in paths if os.path.isdir(os.path.join(path, pth))]
            paths= [pth for pth in dirls if len(os.listdir(os.path.join(path, pth)))>item_n]
        elif kwargs['items'][0]=='=':
            dirls=[pth for pth in paths if os.path.isdir(os.path.join(path, pth))]
            paths= [pth for pth in dirls if len(os.listdir(os.path.join(path, pth)))==item_n]
        elif kwargs['items'][0]=='<':
            dirls=[pth for pth in paths if os.path.isdir(os.path.join(path, pth))]
            paths= [pth for pth in dirls if len(os.listdir(os.path.join(path, pth)))<item_n]
    if 'type' in kwargs:
        if kwargs['type']=='d':
            paths= [pth for pth in paths if os.path.isdir(os.path.join(path, pth))]
        elif kwargs['type']=='f':
            # Should merge with lsfiles
            paths= [pth for pth in paths if os.path.isfile(os.path.join(path, pth))]
        elif kwargs['type']=='l':
            # Should merge with lsfiles
            paths= [pth for pth in paths if os.path.islink(os.path.join(path, pth))]
    if 'ignoreSymlinks' in kwargs:
        if kwargs['ignoreSymlinks']:
            paths= [pth for pth in paths if not os.path.islink(os.path.join(path, pth))]
    if 'onlySymlinks' in kwargs:
        if kwargs['onlySymlinks']:
            paths= [pth for pth in paths if os.path.islink(os.path.join(path, pth))]
    return paths

def lsdir_if(path, **kwargs):
    return lsdir(path, **kwargs)


def jdump(d, fp):  # Just a bit neater to reference
    json.dump(d, open(fp, 'w'))


class BadRequest(Exception):
    pass

def download_file(url, destination_dir=os.getcwd(), destination_filename=None, **kwargs):
    attempt_https=1
    if 'attempt_https' in kwargs:
        attempt_https=kwargs['attempt_https']
    to_print=''
    if url[:7]=='http://':
        to_print+=cl.yellow+'W: Not using HTTPS  '+cl.end
        if attempt_https:
            url='https://'+url[7:]
    if not destination_filename:
        destination_filename = url.split('/')[-1]
    request = requests_get(url, stream=True)
    if request.ok:
        mkdir(destination_dir)
        if 'chunkSize' in kwargs:
            # This method seems required in order to download some files, and might also be faster for very large files
            with open(os.path.join(destination_dir, destination_filename), 'wb') as f:
                for chunk in request.iter_content(chunk_size=kwargs['chunkSize']):
                    f.write(chunk)
        else:
            with open(os.path.join(destination_dir, destination_filename), 'wb') as f:
                shutil.copyfileobj(request.raw, f)
        request.close()
        to_print+='{}{}{} {}/{}'.format(cl.cyan, url, cl.end, destination_dir, destination_filename)
        print(to_print)
    else:
        print(cl.red, 'Bad request', cl.end)
        raise BadRequest('Bad request')


def writeif(path, content, readonly=False, **kwargs):
    if readonly:
        return  # Avoid writes
    # Defaults
    append = False
    # kwargs overrides
    if 'append' in kwargs:
        append = True
    # Main
    if append:
        with open(path, 'a') as f:
            f.write(content)
    else:
        with open(path, 'w') as f:
            f.write(content)


def mkdir(path):
    try:
        os.makedirs(path)
    except FileExistsError:  # Already exists, we don't care about erroring out if it exists.
        return
    except Exception as e:
        print(cl.red, e, cl.end)


def touch(path):
    if not os.path.isfile(path):
        pathlib_Path(path).touch()  # NB: exist_ok=True should also not overwrite the file, but does overwrite metadata


def bytes_to_human_readable(n_bytes):
    if n_bytes < 0:
        return '-' + positive_bytes_to_human_readable(n_bytes)
    else:
        return positive_bytes_to_human_readable(n_bytes)


def positive_bytes_to_human_readable(n_bytes):
    n_bytes = float(n_bytes)
    units = ['bytes', 'KB', 'MB', 'GB', 'TB', 'PB']
    i = 0
    while n_bytes > 1024:
        i += 1
        n_bytes /= 1024.0
    unit = units[i]
    n_bytes = round(n_bytes, 3)
    return str(n_bytes) + ' ' + unit


def memory_usage_raw():
    # This function is entirely the work of Martin Geisler of StackOverflow: https://stackoverflow.com/users/110204/martin-geisler, May 2009
    status = None
    result = {'peak': 0, 'rss': 0}
    try:
        # This will only work on systems such as Linux with a /proc file system
        status = open('/proc/self/status')
        for line in status:
            parts = line.split()
            key = parts[0][2:-1].lower()
            if key in result:
                result[key] = int(parts[1])
    finally:
        if status is not None:
            status.close()
    return result


def memory_usage():
    if os.name == 'posix':  # linux
        d = memory_usage_raw()
        peak = bytes_to_human_readable(d['peak'] * 1024)  # Linux proc is in KB
        rss = bytes_to_human_readable(d['rss'] * 1024)
        return {'peak': peak, 'rss': rss}
    elif use_psutil:
        process = Process(os.getpid())
        return process.memory_info()
    return 'psutil not installed'


def cp(src, dst):
    try:
        shutil.copytree(src, dst)
    except:
        shutil.copy(src, dst)

def free_space(path):
    statvfs = os.statvfs(path)
    return statvfs.f_frsize * statvfs.f_bavail

# Number of free bytes that plebians are allowed to use


file_location = os.path.join(os.path.expanduser('~'), '.config', 'tsr', 'paths.json')
root_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
default_path_settings = os.path.join(root_dir, 'settings')
if os.path.isfile(file_location):
    paths = json.load(open(file_location, 'r'))
else:  # Defaults
    paths = {
        'main': root_dir,
        'posts': os.path.join(root_dir, 'posts'),
        'threads': os.path.join(root_dir, 'threads'),
        'threads_json': os.path.join(root_dir, 'threads.json'),
        'forums': os.path.join(root_dir, 'forums'),
        'users': os.path.join(root_dir, 'users'),
        'users_json': os.path.join(root_dir, 'users.json'),
        'avatars': os.path.join(root_dir, 'avatars'),
        'settings': os.path.join(os.path.expanduser('~'), '.config', 'tsr'),
        'images': os.path.join(os.path.expanduser('~'), 'Pictures', 'tsr'),
        'log': os.path.join(root_dir, 'log.txt')
    }
    cp(default_path_settings, paths['settings'])
    with open(file_location, 'w') as f:
        json.dump(paths, f)

path_variables_str = ' '.join([path for path in paths])
PathObject = namedtuple('path', path_variables_str)
path = PathObject(**paths)
del path_variables_str

if not os.path.isdir(path.settings):
    cp(default_path_settings, path.settings)  # copy default settings into new settings path

for key in paths:  # Make the remaining (empty) folders
    object_path = paths[key]
    if object_path.split('.')[-1] == 'json':
        if not os.path.isfile(object_path):
            open(object_path, 'w').write('{}')
    elif object_path.split('.')[-1] not in ['txt']:
        mkdir(object_path)
    else:
        touch(object_path)



import forumload, soups

class ForumPage(**kwargs):
    def __init__(self, **kwargs):
        if 'browse' in kwargs:
            if kwargs['browse']:
                if ('url' in kwargs)&('forum_id' in kwargs):
                    url=kwargs['ur']
                else:
                    url='https://www.thestudentroom.co.uk/forumdisplay?f='+str(kwargs['forum_id'])
                if 'page_n' in kwargs:
                    page_n=kwargs['page_n']
                else:
                    page_n=1
                
                (self.threads, _, securitytoken)=load_forumpage_threadmeta(kwargs['forum_id'], page_n)



from bs4 import BeautifulSoup
import requests, os.path
from . import os_info # for jdump, path


def main():
    forum_dict = {}

    forum_php = requests.get('https://www.thestudentroom.co.uk/forum.php')
    soup = BeautifulSoup(forum_php.text, 'html.parser')

    categories = soup.find_all('table', {'class': 'forum-category'})
    for category in categories:
        category_title = category.find('th', {'class': 'title'}).find('a').text.strip()
        forum_dict[category_title] = {}
        forums = category.find_all('tr', {'class': 'forum'})
        forums = [f for f in forums if f.find('div', {'class': 'description'})]
        for f in forums:
            title = f.find('a').text.strip()
            forum_id = int(f.find('td', {'class': 'info'})['id'][1:])
            nPosts_tag = f.find('td', {'class': 'count'})
            if nPosts_tag:
                nPosts = int(nPosts_tag.text.strip().replace(',', ''))
            else:
                nPosts = -1
            desc = f.find('div', {'class': 'description'}).text.strip()
            forum_dict[category_title][forum_id] = {
                'title': title,
                'id': forum_id,
                'nPosts': nPosts,
                'desc': desc,
            }
    os_info.jdump(forum_dict, os.path.join(os_info.path.settings, 'forums.json'))



import os, re, subprocess, curses, six, requests
import curses.textpad
import pandas as pd
from math import ceil
from html import unescape
from kitchen.text.display import textual_width_chop
import json # for json.load
# and for json.dump
import matplotlib as plt
from shlex import split as shlex_split

try:
    import nltk # for word_tokenize, pos_tag, PorterStemmer
    use_nltk = True
except ImportError:
    use_nltk = False
from collections import Counter
from time import sleep as time_sleep

from .modules import os_info # for writeif, mkdir, memory_usage, path
from .modules.colours import cl_to_assign
from .modules import general # for make_n_char_long, standardise_datetime, standardise_datetime_str, standardise_date
from .modules import regex, soups
from .modules.forumload import load_forum_database
from .modules.grammar import singularise

class Key:
    ESCAPE = 27
    RETURN = 10
    SPACE = 32


class Port_Source(requests.adapters.HTTPAdapter):
    # Credit: Martijn Pieters https://stackoverflow.com/questions/47202790/python-requests-how-to-specify-port-for-outgoing-traffic
    def __init__(self, port, *args, **kwargs):
        self._source_port = port
        super(Port_Source, self).__init__(*args, **kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = requests.packages.urllib3.poolmanager.PoolManager(
            num_pools=connections, maxsize=maxsize,
            block=block, source_address=('', self._source_port))


def mainbody(screen):  # Over 1000 lines...
    def load_dict_users():
        users = {}
        username_colour_dict = {}  # temporary colour scheme for users without flairs/colours, for easier reading
        username_to_id = json.load(open(os.path.join(os_info.path.settings, 'username_to_id.json')))  # Merge into users.json?
        users = json.load(open(os.path.join(os_info.path.settings, 'users.json')))
        for entry in users:
            if not 'colour' in users[entry]:
                users[entry]['colour'] = 'end'
            if users[entry]['colour'] is None:
                users[entry]['colour'] = 'end'
            username = users[entry]['name']
            if type(username) == 'str':
                username = [username]
            users[entry]['name'] = ls_rm_dupl(username)  # tmp
        return users

    def share_dictionaries():
        soups.settings = settings  # Important to know settings['readonly'] value

    def list_forum_ids(forums):
        string = ''
        for category in forums:
            string += category + '\n'
            for forum in forums[category]:
                string += '  '.join(
                    ['', '', general.make_n_char_long(forums[category][forum]['id'], 4), forums[category][forum]['title']])
                string += '\n            ' + forums[category][forum]['desc']
                string += '\n'
        return string

    def list_commands(commands):
        string = ''
        for entry in commands:
            string += '\n\n' + entry
            string += '\n  ' + commands[entry]['desc'][0]
            print_usages = True
            try:
                string += '\n  ' + commands[entry]['common']['usage']
                print_usages = False
            except:
                pass
            for subentry in commands[entry]['commands']:
                string += '\n  ' + commands[entry]['commands'][subentry]['alts'][0]
                if print_usages:
                    string += '\n    ' + commands[entry]['commands'][subentry]['usage']
        return string

    def session_new():
        session = requests.Session()
        session.headers.update({'User-Agent': settings['user_agent']})
        i = 0
        while i < len(settings['port_sources']):
            try:
                session.mount('https://', Port_Source(settings['port_sources'][i]))
                break
            except:  # Port already in use
                pass
        return session

    def login(securitytoken, **kwargs):
        session = session_new()
        if securitytoken == '':  # Peek-a-boo! Here's some cookies for you!
            session.get('https://www.thestudentroom.co.uk/signin.php')
        credentials = json.load(open(os.path.join(os_info.path.settings, 'credentials.json')))
        if len(credentials) == 0:
            username = prompt_input('Username:    ')
            password = prompt_input('Password:    ')
            userid = prompt_input('User  ID:    ')
            account_credentials = {
                'vb_login_username': username,
                'vb_login_password': password,
                'userid': userid
            }
            credentials.append(account_credentials)
            with open(os.path.join(os_info.path.settings, 'credentials.json'), 'w') as f:
                json.dump(credentials, f)
        elif len(credentials) == 1:
            account_credentials = credentials[0]
        elif len(credentials) > 1:
            if 'username' in kwargs:
                account_username = kwargs['username']
            else:
                account_username = ''
            account_usernames = [x['vb_login_username'] for x in credentials]
            while account_username not in account_usernames:
                account_username = prompt_input('Accounts=' + str(account_usernames))
            account_credentials = [x for x in credentials if x['vb_login_username'] == account_username][0]
        data = dict(
            account_credentials)  # Create an element-wise copy of the dict (rather than one that refers back to the original dict)
        data['do'] = 'login'
        data['securitytoken'] = 'guest'
        data['utcoffset'] = '0'
        data['cookieuser'] = '1'
        url = 'https://www.thestudentroom.co.uk/login.php?do=login'
        request = session.post(url, data)
        securitytoken = soups.get_security_token_from_text(request.text)
        time_sleep(2)
        return session, account_credentials

    def main_init():
        os_info.mkdir(os_info.path.images)

        settings = json.load(open(os.path.join(os_info.path.settings, 'settings.json')))
        commands = json.load(open(os.path.join(os_info.path.settings, 'commands.json')))
        words_colour_dict = json.load(open(os.path.join(os_info.path.settings, 'colours_words.json')))
        domain_colour_dict = json.load(open(os.path.join(os_info.path.settings, 'colours_domains.json')))
        status_message = ''
        if settings['matplotlib']['headless_backend']:
            plt.use('Agg')
            status_message += 'Using Agg backend for matplotlib'

        forums = json.load(open(os.path.join(os_info.path.settings, 'forums.json')))
        if forums == {}:
            import tsr.modules.generate_forums
            tsr.modules.generate_forums.main()

        users = load_dict_users()

        j_meta = {
            'forum_id': 0,
            'number': 0,
            'sorting': ''
        }
        browsing = {
            'logged_in': False,
            'browsing': True,
            'forum_id': 0,
            'sorting': '',
            'thread_id': 0,
            'page_n': 1,
            'cursor': {
                'thread_id': 0,
                'post_id': 0,
                'user': {
                    'name': '',
                    'id': 0,
                },
            },
        }
        securitytoken = ''

        return forums, settings, commands, users, words_colour_dict, domain_colour_dict, j_meta, browsing, securitytoken, status_message

    def strip_textpad(text):
        '''
        The MIT License (MIT)

        Copyright (c) 2015 michael-lazar

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
        '''
        """
        Attempt to intelligently strip excess whitespace from the output of a
        curses textpad.
        """

        if text is None:
            return text

        # Trivial case where the textbox is only one line long.
        if '\n' not in text:
            return text.rstrip()

        # Allow one space at the end of the line. If there is more than one
        # space, assume that a newline operation was intended by the user
        stack, current_line = [], ''
        for line in text.split('\n'):
            if line.endswith('  ') or not line:
                stack.append(current_line + line.rstrip())
                current_line = ''
            else:
                current_line += line
        stack.append(current_line)

        # Prune empty lines at the bottom of the textbox.
        for item in stack[::-1]:
            if not item:
                stack.pop()
            else:
                break

        out = '\n'.join(stack)
        return out

    def add_line(window, text, row=None, col=None, attr=None):
        '''
        The MIT License (MIT)

        Copyright (c) 2015 michael-lazar

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
        '''
        """
        Unicode aware version of curses's built-in addnstr method.

        Safely draws a line of text on the window starting at position
        (row, col). Checks the boundaries of the window and cuts off the text
        if it exceeds the length of the window.
        """

        # The following arg combos must be supported to conform with addnstr
        # (window, text)
        # (window, text, attr)
        # (window, text, row, col)
        # (window, text, row, col, attr)
        cursor_row, cursor_col = window.getyx()
        row = row if row is not None else cursor_row
        col = col if col is not None else cursor_col

        n_cols = w - col - 1
        if n_cols <= 0:
            # Trying to draw outside of the screen bounds
            return

        try:
            text = clean(text, n_cols)
            params = [] if attr is None else [attr]
            window.addstr(row, col, text, *params)
        except (curses.error, ValueError, TypeError) as e:
            # Curses handling of strings with invalid null bytes (b'\00')
            #   python 2: TypeError: "int,int,str"
            #   python 3: ValueError: "embedded null byte"
            _logger.warning('add_line raised an exception')
            _logger.exception(str(e))

    def mainpad_addstr(row, column, content, attr):
        if -1 < row:  # <h-3:
            attr = str2colour[attr]
            mainpad.addstr(row, column, str(content)[:w], attr)

    def sort_by_post_attr(thread_ls, sorting='datetime', reverse=False):
        if len(thread_ls) == 0:
            return thread_ls
        sort_by_user_attr = False
        if sorting[:4] == 'user':
            sorting = sorting[4:]
            sort_by_user_attr = True
        if 'posts' in thread_ls[0]:
            object_to_read_attr_from = sort_by_post_attr(thread_ls[0]['posts'])[
                0]  # Get oldest post - necessary because user may cache the first page after the last
        else:
            object_to_read_attr_from = thread_ls[0]
        if 'datetime' in object_to_read_attr_from:
            sorted_keys_n = 2
            sorted_lastkey_1 = 'datetime'
            sorted_lastkey_2 = 'id'  # In case we have datetimes without times
        else:
            open_pager('Error:\n\n' + str(object_to_read_attr_from))

        # Sanitise thread_ls to remove incomplete threads (eg if scraper is running in background)
        thread_ls = [t for t in thread_ls if 'datetime' in t]

        if sort_by_user_attr:
            if sorted_keys_n == 1:
                thread_ls = sorted(thread_ls, key=lambda x: (x['user'][sorting], x[sorted_lastkey_1]), reverse=reverse)
            elif sorted_keys_n == 2:
                thread_ls = sorted(thread_ls,
                                   key=lambda x: (x['user'][sorting], x[sorted_lastkey_1], x[sorted_lastkey_2]),
                                   reverse=reverse)
        else:
            if sorted_keys_n == 1:
                thread_ls = sorted(thread_ls, key=lambda x: (x[sorting], x[sorted_lastkey_1]),
                                   reverse=reverse)  # Largest values at the top, where the cursor starts
            elif sorted_keys_n == 2:
                thread_ls = sorted(thread_ls, key=lambda x: (x[sorting], x[sorted_lastkey_1], x[sorted_lastkey_2]),
                                   reverse=reverse)
        return thread_ls

    def display_forum(thread_ls, **kwargs):
        mainpad.clear()
        mainpad.bkgd(' ', str2colour['mainpad_bg'])
        if 'sorting' in kwargs:
            sorting = kwargs['sorting']
        else:
            sorting = 'datetime'
        if 'scroll' in kwargs:
            scroll = kwargs['scroll']
        else:
            scroll = 0
        thread_ls = sort_by_post_attr(thread_ls, sorting, True)  # reverse=True
        current_line_y = -scroll  # mainpad_addstr() ignores lines if current_line_y < 0
        for t in thread_ls:
            try:
                datetime = t['datetime']
            except:
                datetime = t['posts'][0]['datetime']
            try:
                title = t['title']
            except:
                title = t['posts'][0]['title']
            try:
                nPosts = t['nPosts']
            except:
                try:
                    nPosts = len(t['posts'])
                except:
                    nPosts = t['nPosts_claimed']
            try:
                reps = t['reps']
            except:
                try:
                    reps = t['posts'][0]['reps']
                except:
                    reps = '?'
            try:
                username = t['user']['name']
            except:
                username = t['posts'][0]['user']['name']
            try:
                username_last = t['last_post_username']
            except:
                username_last = t['posts'][-1]['user']['name']
            datetime, nPosts, reps = str(datetime)[:10], str(nPosts), str(reps)
            mainpad.addstr(current_line_y, 0, str(current_line_y))
            mainpad.addstr(current_line_y, 4, title[:w - 60], str2colour['title'])
            mainpad.addstr(current_line_y, w - 58, datetime, curses.A_BOLD)
            mainpad.attron(str2colour['username'])
            mainpad.addstr(current_line_y, w - 46, username[:20])
            mainpad.addstr(current_line_y, w - 27, username_last[:20])
            mainpad.attroff(str2colour['username'])
            mainpad.addstr(current_line_y, w - 10, nPosts, str2colour['info'])
            mainpad.addstr(current_line_y, w - 4, reps, str2colour['reps'])
            current_line_y += 1
        if current_line_y > 0:
            post_start_lines = range(
                current_line_y)  # There is (currently) one thread entry per line, so use range, and we don't want to error out when
        else:
            post_start_lines = [0]
        return current_line_y, post_start_lines, thread_ls

    def display_posts(posts_ls, sorting='datetime'):
        mainpad.clear()
        current_line_y = 0
        post_start_lines = []
        posts_ls = sort_by_post_attr(posts_ls, sorting, True)  # reverse=True
        for post in posts_ls:
            post_start_lines.append(current_line_y)
            userid = post['user']['id']
            mainpad.addstr(current_line_y, 0, '-' * w, str2colour['faint'])  # Faint line to help viewing
            mainpad.addstr(current_line_y, 0, post['user']['name'], str2colour['username'])
            mainpad.addstr(current_line_y, 20, str(post['user']['id']), str2colour['userid'])
            try:
                mainpad.addstr(current_line_y, 30, str(post['user']['reps']), str2colour['reps'])
            except:  # User has hidden their reputation bar
                pass
            next_x = 33
            try:
                mainpad.addstr(current_line_y, 30, users[userid]['flair'], str2colour['userflair'])
                next_x = 30 + len(users[userid]['flair']) + 1
            except:
                pass
            mainpad.addstr(current_line_y, next_x, str(post['datetime']), str2colour['datetime'])
            if post['reps'] > 0:
                mainpad.addstr(current_line_y, w - 4, str(post['reps']), str2colour['reps'])

            for content in post['contents']:
                c = content['content']
                if content['type'] == 'spoiler':
                    c_normal_attr = '@@@spoiler~~~' + c.replace('\n', '\n~')
                elif content['type'] == 'quote':
                    c_normal_attr = '@@@quote~~~~~'
                    c = '>>' + content['user']['name'] + '    ' + str(content['post']['id']) + '\n>' + c.replace('\n',
                                                                                                                 '\n>')
                elif content['type'] == 'text':
                    c_normal_attr = '@@@default~~~'
                if len(c) > 1:
                    c = c.replace('@@@default~~~',
                                  c_normal_attr)  # In case we've defined locations of default colours before
                    c = c_normal_attr + c

                    usernames = set([re.escape(post['user']['name']) for post in
                                     posts_ls])  # re.escape means special characters dont affect our regex later
                    usernames_regex = '|'.join(usernames)
                    try:
                        c = re.sub(regex.url2, r'{0}\1{1}'.format('@@@url~~~~~~~', c_normal_attr), c,
                                   flags=re.MULTILINE)  # Colour urls
                        c = re.sub(regex.image, r'{0}\1{1}'.format('@@@image~~~~~', c_normal_attr), c,
                                   flags=re.MULTILINE)  # Colour images (inc. image filenames in urls) purple
                        c = re.sub(r'({0})'.format(usernames_regex), r'{0}\1{1}'.format('@@@username~~', c_normal_attr),
                                   c, flags=re.MULTILINE)  # Colour suspected usernames (in text)
                        c = regex.parse_custom_colours(c)
                    except:
                        pass

                    c_coloured = [[], []]
                    c_colours = []
                    c_split = c.split('@@@')
                    for i in range(len(c_split)):
                        cc = c_split[i]
                        attr = cc[:10].replace('~', '')
                        open('/tmp/usetwoleft', 'a').write('\n\n\nAttr: <' + attr + '>')
                        if attr == 'useTwoLeft':
                            open('/tmp/usetwoleft', 'a').write('\nAttr: <useTwoLeft> confirmed')
                            open('/tmp/usetwoleft', 'a').write('\nlist: ' + str(c_colours[1]))
                            if i > 1:
                                attr = c_colours[i - 2]
                                open('/tmp/usetwoleft', 'a').write('\nRepl: <' + str(attr) + '>')
                            else:
                                attr = str2colour['default']
                                open('/tmp/usetwoleft', 'a').write('\nRepl: <default> (bcos i<2)')
                        elif attr == '':
                            attr = curses.A_NORMAL
                            open('/tmp/usetwoleft', 'a').write('\nRepl: <A_NORMAL> (bcos blank)')
                        else:
                            attr = str2colour[attr]
                            open('/tmp/usetwoleft', 'a').write('\nSmoothly done')
                        c_coloured[0] += cc[10:]
                        c_colours.append(attr)
                        for i in range(len(cc) - 10):
                            c_coloured[1].append(attr)
                    current_line_x = 0
                    for i in range(len(c_coloured[0])):
                        char = c_coloured[0][i]
                        attr = c_coloured[1][i]
                        if i % w == 0:
                            current_line_y += 1
                            current_line_x = 0
                        mainpad.addstr(current_line_y, current_line_x, char, attr)
                        if char == '\n':
                            current_line_y += 1
                            current_line_x = 0
                        else:
                            current_line_x += 1
            current_line_y += 2
        return current_line_y, post_start_lines, posts_ls  # In case we want to append things to the mainpad

    def display_tuple_ls(ls):
        if type(ls) != type([]):
            raise ValueError('display_tuple_ls requires LIST argument')
        if len(ls) == 0:
            open_pager('display_tuple_ls: No results\n\n' + str(ls))
            return 0  # Not necessarily a bug if we get no results
        mainpad.clear()
        current_line_y = 0
        max_w_rows_ls = [max([len(str(tpl[i])) for tpl in ls]) for i in
                         range(len(ls[0]))]  # Maximum size of elements in each column
        # So we can easily say that offset of Nth element = max_length of (N-1)th element
        colours = [curses.color_pair(i) for i in range(len(max_w_rows_ls))]  # For alternating colours
        for tpl in ls:
            x_offset = 0
            for i in range(len(ls[0])):
                mainpad.addstr(current_line_y, x_offset, str(tpl[i]), colours[i])
                x_offset += max_w_rows_ls[i] + 1
            current_line_y += 1
        return current_line_y

    def get_attr(window, y, x):
        coord_value = window.inch(y, x)
        ch = coord_value & 255  # first 8 bits (2^8=256) are char
        attr_value = int((int(coord_value) - int(ch)) / 256)
        return attr_value

    def clean(string, n_cols=None):  # browse t 5316512    5316100
        '''
        The MIT License (MIT)

        Copyright (c) 2015 michael-lazar

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
        '''
        """
        Required reading!
            http://nedbatchelder.com/text/unipain.html

        Python 2 input string will be a unicode type (unicode code points).
        Curses will accept unicode if all of the points are in the ascii range.
        However, if any of the code points are not valid ascii curses will
        throw a UnicodeEncodeError: 'ascii' codec can't encode character,
        ordinal not in range(128). If we encode the unicode to a utf-8 byte
        string and pass that to curses, it will render correctly.

        Python 3 input string will be a string type (unicode code points).
        Curses will accept that in all cases. However, the n character count in
        addnstr will not be correct. If code points are passed to addnstr,
        curses will treat each code point as one character and will not account
        for wide characters. If utf-8 is passed in, addnstr will treat each
        'byte' as a single character.

        Reddit's api sometimes chokes and double-encodes some html characters
        Praw handles the initial decoding, but we need to do a second pass
        just to make sure. See https://github.com/michael-lazar/rtv/issues/96

        Example:
            &amp;amp; -> returned directly from reddit's api
            &amp;     -> returned after PRAW decodes the html characters
            &         -> returned after our second pass, this is the true value
        """

        if n_cols is not None and n_cols <= 0:
            return ''

        if isinstance(string, six.text_type):
            string = unescape(string)

        if False:  # self.config['ascii']:
            if isinstance(string, six.binary_type):
                string = string.decode('utf-8')
            string = string.encode('ascii', 'replace')
            return string[:n_cols] if n_cols else string
        else:
            if n_cols:
                string = textual_width_chop(string, n_cols)
            if isinstance(string, six.text_type):
                string = string.encode('utf-8')
            return string

    def curs_set(n):
        try:
            curses.curs_set(n)
        except:
            pass

    def text_input(window=False, allow_resize=False):
        '''
        The MIT License (MIT)

        Copyright (c) 2015 michael-lazar

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
        '''
        """
        Transform a window into a text box that will accept user input and loop
        until an escape sequence is entered.

        If the escape key (27) is pressed, cancel the textbox and return None.
        Otherwise, the textbox will wait until it is full (^j, or a new line is
        entered on the bottom line) or the BEL key (^g) is pressed.
        """
        if not window:
            window = curses.newwin(h - 3, w, 2, 0)
        window.clear()

        # Set cursor mode to 1 because 2 doesn't display on some terminals
        curs_set(1)

        # Keep insert_mode off to avoid the recursion error described here
        # http://bugs.python.org/issue13051
        textbox = curses.textpad.Textbox(window)
        textbox.stripspaces = 0

        def validate(ch):
            "Filters characters for special key sequences"
            if ch == Key.ESCAPE:
                raise exceptions.EscapeInterrupt()
            if (not allow_resize) and (ch == curses.KEY_RESIZE):
                raise exceptions.EscapeInterrupt()
            # Fix backspace for iterm
            if ch == curses.ascii.DEL:
                ch = curses.KEY_BACKSPACE
            return ch

        out = textbox.edit(validate=validate)
        if isinstance(out, six.binary_type):
            out = out.decode('utf-8')

        curs_set(0)
        return strip_textpad(out)

    def suspend():
        '''
        The MIT License (MIT)

        Copyright (c) 2015 michael-lazar

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
        '''
        """
        Suspend curses in order to open another subprocess in the terminal.
        """

        try:
            curses.endwin()
            yield
        finally:
            curses.doupdate()

    def open_pager(data, wrap=None):
        '''
        The MIT License (MIT)

        Copyright (c) 2015 michael-lazar

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
        '''
        """
        View a long block of text using the system's default pager.

        The data string will be piped directly to the pager.
        """

        pager = os.getenv('PAGER') or 'less'
        command = shlex_split(pager)

        if wrap:
            data_lines = content.Content.wrap_text(data, wrap)
            data = '\n'.join(data_lines)

        try:
            curses.endwin()
            p = subprocess.Popen(command, stdin=subprocess.PIPE)
            try:
                p.communicate(data.encode('utf-8'))
            except KeyboardInterrupt:
                p.terminate()
            curses.doupdate()
        except OSError as e:
            screen.show_notification('Could not open pager %s' % pager)

    def prompt_input(prompt, key=False):
        '''
        The MIT License (MIT)

        Copyright (c) 2015 michael-lazar

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
        '''
        """
        Display a text prompt at the bottom of the screen.

        Params:
            prompt (string): Text prompt that will be displayed
            key (bool): If true, grab a single keystroke instead of a full
                        string. This can be faster than pressing enter for
                        single key prompts (e.g. y/n?)
        """
        v_offset, h_offset = screen.getbegyx()
        prompt = clean(prompt, w - 1)

        # Create a new window to draw the text at the bottom of the screen,
        # so we can erase it when we're done.
        s_row = v_offset + h - 1
        s_col = h_offset
        prompt_win = curses.newwin(1, len(prompt) + 1, s_row, s_col)
        prompt_win.bkgd(' ', str2colour['prompt'])
        add_line(prompt_win, prompt)
        prompt_win.refresh()

        # Create a separate window for text input
        s_col = h_offset + len(prompt)
        input_win = curses.newwin(1, w - len(prompt), s_row, s_col)
        input_win.bkgd(str(' '), str2colour['prompt'])
        input_win.refresh()

        if key:
            curs_set(1)
            ch = screen.getch()
            # We can't convert the character to unicode, because it may return
            # Invalid values for keys that don't map to unicode characters,
            # e.g. F1
            text = ch if ch != Key.ESCAPE else None
            curs_set(0)
        else:
            text = text_input(input_win)

        prompt_win.clear()
        input_win.clear()
        del prompt_win
        del input_win
        screen.touchwin()
        screen.refresh()

        return text

    def session_post(url, data, **kwargs):
        global browsing  # Global variables are bad. But pain to do otherwise.
        if 'logout' in kwargs:
            if kwargs['logout']:
                session = session_new()
        if not browsing['logged_in']:
            do_login = True
            if kwargs['login'] in kwargs:
                if not kwargs['login']:
                    do_login = False
            if do_login:
                session, account_credentials = login(securitytoken)
                browsing['logged_in'] = True
        securitytoken = soups.parse_page(page_type='register', session=session)
        session.post(url, data)
        return session

    def post_comment(session, **kwargs):
        if not browsing['logged_in']:
            session, account_credentials = login(securitytoken)
            browsing['logged_in'] = True
        if 'reply_to' in kwargs:
            post_id = kwargs['reply_to']
            post = [x for x in posts_list if x['id'] == post_id][0]
            username = post['user']['name']
            content = '\n'.join([x['content'] for x in post['contents'] if x['type'] == 'text'])  # Don't quote quotes
            prefill = '[QUOTE={};{}]{}[/QUOTE]\n'.format(username, str(post_id), content)
            open_pager(prefill)
        else:
            open_pager('''CTRL+G to post''')
        data = {
            'fromquickreply': '1',
            'securitytoken': securitytoken,
            'do': 'postreply',
            'p': 'who cares',
            'specifiedpost': '0',
            'parseurl': '1',
            'loggedinuser': str(account_credentials['userid']),
            'postroute': 'thread/bottom',
            'message': text_input(),
            'signature': '1',
            'sbutton': 'Submit Reply',
        }
        if 'thread_id' in kwargs:
            if securitytoken == 'guest':
                (_, data['securitytoken']) = soups.load_thread_page(kwargs['thread_id'], 1, session=session)
            url = 'https://www.thestudentroom.co.uk/newreply.php?do=postreply&t=' + str(
                kwargs['thread_id'])  # We are assuming we've run 'browse $THREAD' before this
            data['t'] = str(kwargs['thread_id'])

        session.post(url, data)
        return 'posted'

    screen.clear()
    screen.refresh()

    curses.start_color()
    curses.use_default_colors()

    (forums, settings, commands, users, words_colour_dict, domain_colour_dict, j_meta, browsing, securitytoken,
     status_message) = main_init()

    theme_to_bg = {
        'dark': curses.COLOR_BLACK,
        'paper': curses.COLOR_WHITE,
    }
    mainpad_bg = theme_to_bg[settings['theme']]
    if mainpad_bg == curses.COLOR_BLACK:
        mainpad_bg_opposite = curses.COLOR_WHITE
    elif mainpad_bg == curses.COLOR_WHITE:
        mainpad_bg_opposite = curses.COLOR_BLACK

    curses.init_pair(0, 0, mainpad_bg)
    curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLUE)  # Title: white fg, black bg
    curses.init_pair(2, curses.COLOR_CYAN, mainpad_bg)
    curses.init_pair(3, curses.COLOR_GREEN, mainpad_bg)
    curses.init_pair(4, curses.COLOR_YELLOW, mainpad_bg)  # Warning
    curses.init_pair(5, curses.COLOR_RED, mainpad_bg)
    curses.init_pair(6, -1, curses.COLOR_MAGENTA)  # Command
    curses.init_pair(7, curses.COLOR_BLACK, curses.COLOR_CYAN)
    curses.init_pair(8, curses.COLOR_MAGENTA, mainpad_bg)  # Command
    curses.init_pair(9, curses.COLOR_BLUE, mainpad_bg)  # Command
    curses.init_pair(10, mainpad_bg_opposite, mainpad_bg)
    curses.init_pair(11, curses.COLOR_WHITE, curses.COLOR_BLACK)
    curses.init_pair(12, curses.COLOR_RED, mainpad_bg)
    if mainpad_bg == curses.COLOR_BLACK:
        curses.init_pair(13, curses.COLOR_BLUE, mainpad_bg)
    elif mainpad_bg == curses.COLOR_WHITE:
        curses.init_pair(13, curses.COLOR_YELLOW, mainpad_bg)
    curses.init_pair(14, curses.COLOR_RED, mainpad_bg_opposite)
    curses.init_pair(15, curses.COLOR_BLACK, curses.COLOR_RED)
    curses.init_pair(16, curses.COLOR_WHITE, mainpad_bg)
    curses.init_pair(17, curses.COLOR_BLACK, mainpad_bg)

    str2colour = {
        'default': curses.color_pair(0),
        'end': curses.color_pair(0),  # Old default

        'header': curses.color_pair(1),
        'title': curses.color_pair(4),
        'info': curses.color_pair(2),
        'success': curses.color_pair(3),
        'warning': curses.color_pair(4),
        'refuse': curses.color_pair(5),
        'error': curses.color_pair(5),
        'command': curses.color_pair(6),
        'url': curses.color_pair(2),
        'username': curses.color_pair(4),
        'userid': curses.color_pair(2),
        'userflair': curses.color_pair(1),
        '@username': curses.color_pair(1),  # Tag used to link users, similar to /u/Username
        'status': curses.color_pair(7),
        'prompt': curses.color_pair(7),
        'reps': curses.color_pair(3),
        'date': curses.color_pair(2),
        'datetime': curses.color_pair(2),
        'quote': curses.color_pair(8),
        'spoiler': curses.color_pair(9),
        'purple': curses.color_pair(8),
        'image': curses.color_pair(2),
        'mainpad_bg': curses.color_pair(10),
        'grep': curses.color_pair(12),
        'faint': curses.color_pair(13),
        'r_opposite': curses.color_pair(14),

        'white': curses.color_pair(16),
        'grey': curses.color_pair(17),
        'lgrey': curses.color_pair(17),
        'black': curses.color_pair(17),

        'purple': curses.color_pair(8),
        'lpurple': curses.color_pair(8),
        'dpurple': curses.color_pair(8),
        'red': curses.color_pair(5),
        'lred': curses.color_pair(5),
        'dred': curses.color_pair(5),
        'yellow': curses.color_pair(4),
        'lyellow': curses.color_pair(4),
        'dyellow': curses.color_pair(4),
        'orange': curses.color_pair(4),
        'brown': curses.color_pair(4),
        'green': curses.color_pair(3),
        'lgreen': curses.color_pair(3),
        'dgreen': curses.color_pair(3),
        'cyan': curses.color_pair(2),
        'lcyan': curses.color_pair(2),
        'dcyan': curses.color_pair(2),
        'blue': curses.color_pair(9),
        'lblue': curses.color_pair(9),
        'dblue': curses.color_pair(9),

        'white_on_black': curses.color_pair(11),
        'black_red': curses.color_pair(15),

        'bold': curses.A_BOLD,
        'underlined': curses.A_UNDERLINE,
        'blink': curses.A_BLINK,
        'dim': curses.A_DIM,
        'standout': curses.A_STANDOUT,
    }

    h, w = screen.getmaxyx()

    mainpad_size = 0
    while True:
        try:
            mainpad_size += 1
            mainpad = curses.newpad(mainpad_size * 1000,
                                    w)  # curses.newwin(h-3,w, 2,0) # #curses.newpad(mainpad_size*1000,w) # Main pad for displaying threads,posts,etc, easy scrolling
            del mainpad
        except:
            mainpad_size -= 1
            mainpad = curses.newpad(mainpad_size * 1000,
                                    w)  # curses.newwin(h-3,w, 2,0) #curses.newpad(mainpad_size*1000,w)
            break

    mainpad.scrollok(True)
    mainpad.bkgd(str(' '), str2colour['mainpad_bg'])
    username_to_id = {}
    j = {}
    command_last = ''

    current_line_y = 0

    key = 0
    cursor_x, cursor_y = 0, 0
    commandstr = ''
    command = ''
    command_last = ''
    commandstr_last = ''
    arguments_original = []
    last_object_displayed = ''
    scroll = 0
    mainpad_scroll = 0
    last_attr = 0
    height_of_current_displayed_obj = 2
    status_message = 'Loaded'
    commanded = False
    session = session_new()  # Pre-login, don't really need a session, but it helps us a) define our custom user agent, and b) makes the following script neater (no need to account for not having a session defined)
    persistant_bars = {
        'title': {
            'content': 'TITLE',
            'x': 0,
            'y': 0,
            'attr': [str2colour['header'], curses.A_BOLD],
        },
        'subtitle': {
            'content': 'SUBTITLE',
            'x': 0,
            'y': 1,
            'attr': [str2colour['header']],
        },
        'status': {
            'content': '{11} | last_object_displayed:{12} | Mainbar_h:{7}k | Last key:{3} | Last command:{4} | RAM:{2} | j_meta:{5} | token:{6}'.format(
                cursor_x, cursor_y, str(os_info.memory_usage()['rss']), key, commandstr_last, str(j_meta), securitytoken,
                str(mainpad_size), command_last, str(mainpad_scroll), str(h), str(status_message),
                last_object_displayed),
            'x': 0,
            'y': h - 1,
            'attr': [str2colour['status']],
        },
    }
    scroll_lines = [0]
    while key != ord('q'):
        # screen.clear()
        h, w = screen.getmaxyx()
        cursor_x_last, cursor_y_last = cursor_x, cursor_y
        if key == curses.KEY_DOWN:
            if cursor_y == scroll_lines[-1]:  # At the end of the page
                if (last_object_displayed == 'thread') & (browsing['browsing']):
                    commandstr = 'browse t ' + str(browsing['thread_id']) + ' ' + str(
                        browsing['page_n'] + 1)  # Load next page
                    commanded = True
            else:
                cursor_y_new = [row for row in scroll_lines if row > cursor_y] + [h]
                cursor_y_new = cursor_y_new[0]
                cursor_y_diff = cursor_y_new - cursor_y
                cursor_y = cursor_y_new
                if mainpad_scroll < height_of_current_displayed_obj:
                    mainpad_scroll += cursor_y_diff
        elif key == curses.KEY_UP:
            cursor_y_new = [0] + [row for row in scroll_lines if row < cursor_y]
            cursor_y_new = cursor_y_new[-1]
            cursor_y_diff = cursor_y_new - cursor_y
            cursor_y = cursor_y_new
            if cursor_y - mainpad_scroll < 1:
                mainpad_scroll += cursor_y_diff
        elif key == curses.KEY_RIGHT:
            if last_object_displayed == 'forum':
                if browsing['browsing']:
                    commandstr = 'browse ' + str(cursor_y)
                    commanded = True
                else:
                    commandstr = str(cursor_y)  # Index of highlighted thread
                    commanded = True
            else:
                cursor_x += 1
        elif key == curses.KEY_LEFT:
            if last_object_displayed != 'forum':
                commandstr = 'back'  # Index of highlighted thread
                commanded = True
            else:
                cursor_x -= 1
        elif key == ord(':'):
            command_prompt = 'Command :    '
            commandstr = prompt_input(command_prompt)
            commanded = True
        elif key == ord('r'):  # Reply
            if last_object_displayed == 'thread':
                status_message = post_comment(session, thread_id=browsing['thread_id'],
                                              reply_to=browsing['cursor']['post_id'])
            elif last_object_displayed == 'forum':  # So we want to reply to a thread, i.e. post a normal comment
                status_message = post_comment(session, thread_id=browsing['thread_id'])
        elif key == ord('R'):  # Award rep
            commandstr = 'rep ' + str(browsing['cursor']['post_id'])
            commanded = True

        if key in [curses.KEY_UP, curses.KEY_DOWN]:
            if last_object_displayed == 'thread':
                browsing['cursor']['post_id'] = posts_list[scroll_lines.index(cursor_y)]['id']
            elif last_object_displayed == 'forum':
                try:
                    browsing['cursor']['thread_id'] = thread_meta_list[scroll_lines.index(cursor_y)]['id']
                except:
                    if len(j) > 0:
                        browsing['cursor']['thread_id'] = j[min(scroll_lines.index(cursor_y), max(0, len(j) - 1))]['id']
                    else:
                        browsing['cursor']['thread_id'] = 0
                # Thread ID of the thread underneath the cursor
                # Actually would be better to define this alongside scroll_lines in a dictionary, on forum_load and thread_load

        commandstr_last = commandstr
        persistant_bars['status'][
            'content'] = '{11} | last_object_displayed:{12} | Mainbar_h:{7}k | Cursor:{3} | Last command:{4} | RAM:{2} | j_meta:{5} | token:{6}'.format(
            cursor_x, cursor_y, str(os_info.memory_usage()['rss']), str(browsing['cursor']), commandstr_last, str(j_meta),
            securitytoken, str(mainpad_size), command_last, str(mainpad_scroll), str(h), str(status_message),
            last_object_displayed)
        command_last = command
        arguments_last = arguments_original
        if commanded:
            arguments = shlex_split(commandstr)
            arguments_original = arguments
            command = arguments[0]
            arguments = arguments[1:]
            mainpad_scroll, cursor_y = 0, 0
            commanded = False
        else:
            command = ''
            arguments = []
            arguments_original = arguments = []
        cursor_x = min(max(0, cursor_x), w - 1)
        cursor_y = min(max(0, cursor_y), height_of_current_displayed_obj - 1)

        mainpad_scroll = min(max(0, mainpad_scroll), height_of_current_displayed_obj - 1)

        try:
            mainpad.chgat(cursor_y_last, cursor_x_last, 1, curses.color_pair(last_attr))  # Undo previous highlight
            last_attr = get_attr(mainpad, cursor_y, cursor_x)
            if mainpad_bg == curses.COLOR_BLACK:
                mainpad.chgat(cursor_y, cursor_x, 1, curses.A_REVERSE)  # Highlight 1 char at cursor location
            elif mainpad_bg == curses.COLOR_WHITE:
                mainpad.chgat(cursor_y, cursor_x, 1,
                              str2colour['white_on_black'])  # Highlight 1 char at cursor location
        except:  # We haven't drawn anything on the display, hence cannot change its colour
            pass

        for bar in persistant_bars:
            for attr in persistant_bars[bar]['attr']:
                screen.attron(attr)
            x = persistant_bars[bar]['x']
            y = persistant_bars[bar]['y']
            content = persistant_bars[bar]['content'][:w - 1]  # So it doesn't overflow
            while len(content) < w - 1:
                content += ' '
            screen.addstr(y, x, content)
            for attr in persistant_bars[bar]['attr']:
                screen.attroff(attr)

        mainpad.refresh(mainpad_scroll, 0, 2, 0, h - 1, w)
        screen.refresh()

        key = screen.getch()

        share_dictionaries()

        for entry in commands['threaddisplay']['commands']:
            if command in ['grep', 'egrep']:
                pass  # temporary
            elif command in commands['threaddisplay']['commands'][entry]['alts']:
                attr_value = arguments[0]
                arguments = arguments[1:]
                try:
                    number_to_print = int(arguments[1])
                    arguments = arguments[1:]
                except:
                    number_to_print = 50
                try:
                    sorting = arguments[0]
                except:
                    sorting = 'datetime'
                posts_list = []
                post_ids = []
                attr_is_user_attr = False
                for thread in j:
                    for post in thread['posts']:
                        if entry in post:
                            if post[entry] == attr_value:
                                posts_list.append(post)
                        elif (entry[:4] == 'user'):
                            attr = entry[4:]
                            attr_is_user_attr = True
                            if attr in post['user']:
                                if post['user'][attr] == attr_value:
                                    posts_list.append(post)
                try:
                    posts_list = sorted(posts_list, key=lambda x: x['user'][sorting], reverse=True)
                except:
                    posts_list = sorted(posts_list, key=lambda x: x[sorting], reverse=True)
                height_of_current_displayed_obj, scroll_lines, posts_list = display_posts(posts_list)
        if command == 'total':
            current_line_y = 2
            if len(arguments) == 0:
                open_pager('''Rejected

                Requires an argument

                Attributes include
                    username userid date reps domain imghost

                eg
                    total username
                    total hour 50''')
            else:
                filter_attr = arguments[0]
                arguments = arguments[1:]
                attr_is_user_attr = False
                if filter_attr[:4] == 'user':
                    filter_attr = filter_attr[4:]
                    attr_is_user_attr = True
                try:
                    number_to_print = int(arguments[0])
                    arguments = arguments[1:]
                except:
                    number_to_print = 50
                try:
                    min_occurances = int(arguments[1])
                    arguments = arguments[1:]
                except:
                    min_occurances = 1
                try:
                    sort_by = arguments[0]
                except:
                    sort_by = 'count'

                attrs = Counter([])
                datetimes = []
                for thread in j:
                    attrs_to_append = []  # Clear it each time to avoid memory explosion in the case of words. Listing every word from every post from every thread is bad idea. Counter is probably more easy on RAM.
                    thread_new = {x: thread[x] for x in thread if x != 'posts'}
                    thread_new['posts'] = []
                    for post in thread['posts']:
                        if attr_is_user_attr:
                            if filter_attr in post['user']:
                                append_this = True
                                to_append = post['user'][filter_attr]
                        else:
                            if filter_attr in ['all', 'time', 'datetime', 'minute', 'hour', 'day', 'month', 'year']:
                                to_append = general.standardise_datetime_str(post['datetime'])
                                append_this = False
                                datetimes.append(to_append)
                            elif filter_attr in post:
                                to_append = post[filter_attr]
                                append_this = True
                            elif filter_attr in ['word', 'words']:
                                content = '\n'.join([x['content'] for x in post['contents'] if x['type'] != 'quote'])
                                to_append = regex.ls_words(
                                    content)  # WARNING: This can eat up a lot of memory, probably a bad idea to use on an entire forum consisting of 500 threads.
                                append_this = True
                            elif filter_attr in ['domain', 'domains']:
                                content = '\n'.join([x['content'] for x in post['contents'] if x['type'] != 'quote'])
                                to_append = regex.findall_domains(content)
                                append_this = True
                            elif filter_attr in ['imghost', 'imagehost', 'imghosts', 'imagehosts']:
                                content = '\n'.join([x['content'] for x in post['contents'] if x['type'] != 'quote'])
                                to_append = regex.findall_domains(content, only_image_hosts=True)
                                append_this = True
                        if append_this:
                            if type(to_append).__name__ in ['list']:
                                for child in to_append:
                                    attrs_to_append.append(child)
                            else:
                                attrs_to_append.append(to_append)
                    attrs = attrs + Counter(attrs_to_append)
                attrs = [(x, attrs[x]) for x in attrs if attrs[x] > min_occurances]
                if len(attrs) == 0:
                    status_message = 'No results'
                if filter_attr in ['all', 'time', 'date', 'minute', 'hour', 'day', 'month', 'year']:
                    df = pd.DataFrame()
                    df['datetime'] = pd.to_datetime(datetimes, format='%Y-%m-%d %H:%M:%S')
                    if filter_attr in ['all', 'date']:  # Has to come first, before df=df.groupby
                        df = df.groupby(df['datetime'].dt.normalize())
                        df.count().plot(figsize=(settings['matplotlib']['w'] / 100, settings['matplotlib']['h'] / 100))
                    else:
                        if filter_attr in ['all', 'time']:
                            df = df.groupby(df['datetime'].dt.time)
                        if filter_attr in ['all', 'minute']:
                            df = df.groupby(df['datetime'].dt.minute)
                        if filter_attr in ['all', 'hour']:
                            df = df.groupby(df['datetime'].dt.hour)
                        if filter_attr in ['all', 'day']:
                            df = df.groupby(df['datetime'].dt.weekday)
                        if filter_attr in ['all', 'month']:
                            df = df.groupby(df['datetime'].dt.month)
                        if filter_attr in ['all', 'year']:
                            df = df.groupby(df['datetime'].dt.year)
                        df.count().plot(kind='bar',
                                        figsize=(settings['matplotlib']['w'] / 100, settings['matplotlib']['h'] / 100))
                    if settings['matplotlib']['headless_backend']:
                        previous_plots = [x for x in os.listdir(os_info.path.main) if x.split('.')[-1] == 'png']
                        path_plot = os.path.join(os_info.path.images, filter_attr + '.png')
                        plt.pyplot.figure(
                            figsize=(settings['matplotlib_w'] / 100, settings['matplotlib_h'] / 100))
                        plt.pyplot.savefig(
                            path_plot)  # dpi of 1 doesnt work well with fonts etc, 100 is near standard dpi of 96
                        status_message = 'Saved as: ' + path_plot
                        plt.pyplot.close()
                    else:
                        plt.pyplot.figure(
                            figsize=(settings['matplotlib']['w'] / 100, settings['matplotlib']['h'] / 100))
                        plt.pyplot.show()
                        plt.pyplot.close()
                else:
                    attrs = sorted(attrs, key=lambda x: x[1], reverse=True)
                    attrs = attrs[:number_to_print]
                    height_of_current_displayed_obj = display_tuple_ls(attrs)
        elif command == 'help':
            if len(arguments) == 0:
                open_pager(list_forum_ids(forums) + '\n\n' + list_commands(
                    commands) + '\n\n' + 'If errors, check ~/.config/tsr/paths.json and $settingsPath/settings.json')
            elif len(arguments) == 1:
                try:
                    open_pager(commands[arguments[0]]['common']['help'])
                except:
                    open_pager(commands[arguments[0]]['desc'][-1])
            elif len(arguments) == 2:
                open_pager(commands[arguments[0]]['commands'][arguments[1]]['help'])
            open_pager('''
            List of commands:
            Loading       load
            Settings      setting flair colour dict
            Analytics     total byday
            Filters       filter grep egrep
            Browsing      browse url browser
            If logged in  post report rep
            Misc          back t
            ''')
        elif command == 'back':
            persistant_bars['subtitle']['content'] = ''
            last_object_displayed = 'forum'
            if browsing['browsing']:
                height_of_current_displayed_obj, scroll_lines, thread_meta_list = display_forum(thread_meta_list)
            else:  # So we can load more meta information when displaying the forum-like object
                try:
                    number_of_threads = int(arguments[0])
                    try:
                        sorting = arguments[1]
                    except:
                        sorting = 'datetime'
                except:
                    number_of_threads = 500
                    try:
                        sorting = arguments[0]
                    except:
                        sorting = 'datetime'
                height_of_current_displayed_obj, scroll_lines, j = display_forum(j, sorting=sorting)
        elif command == 'setting':
            if len(arguments) == 2:
                key = arguments[0]
                if key in settings:
                    value = arguments[1]
                    if value in ['false', 'False']:
                        value = False
                    elif value in ['true', 'True']:
                        value = True
                    else:
                        try:
                            value = int(value)
                        except:
                            try:
                                value = float(value)
                            except:
                                pass
                    settings[key] = value
            else:
                toprint = '''
                Error:  Requires two arguments

                Usage:  setting [variable] [value]

                Eg:
                    setting readonly true       (to avoid writing anything to disk - be it new settings, page caches, or logs)

                Variables are:'''
                for key in settings:
                    toprint += '\n                    {}'.format(str(key))
                open_pager(toprint)
        elif command in ['tag', 'flair', 'colour']:
            username = arguments[0]
            if command == 'colour':
                if arguments[1] not in colour:
                    toprint = '''
                    Error: Colour not available

                    Available colours are:'''
                    for key in colour:
                        toprint += '\n' + colour[key] + '   ' + key
                    open_pager(toprint)
            try:
                userid = str(username_to_id[username])
                username_is_name = True
                proceed = True
            except:
                username_is_name = False
                try:
                    userid = str(int(username))
                    proceed = True
                except:
                    userid = ''  # Prevent us from using the last value of userid, ie the wrong user
                    toprint = '''
                    Username not known. Try user_id instead.

                    Known usernames are:
                    ''' + str('\n'.join(username_to_id))
                    open_pager(toprint)
                    proceed = False
            if proceed:
                try:
                    if username_is_name:
                        try:
                            if username not in users[userid]['name']:
                                users[userid]['name'].append(username)
                        except:
                            users[userid]['name'] = [username]
                    users[userid][command] = arguments[1]
                except:
                    users[userid] = {'name': [username], command: arguments[1]}
        elif command == 'dict':
            command = arguments[0]
            arguments = arguments[1:]
            if command == 'save':
                if arguments[0] == 'all':
                    arguments = commands['dictionary']['commands']['save']['help']['dictionary']
                if 'settings' in arguments:
                    with open(os.path.join(os_info.path.settings, 'settings.json'), 'w') as f:
                        json.dump(settings, f)
                    status_message = 'Saved settings.json'
                if 'users' in arguments:
                    with open(os.path.join(os_info.path.settings, 'users.json'), 'w') as f:
                        json.dump(users, f)
                    status_message = 'Saved users.json'
                if 'domains' in arguments:
                    with open(os.path.join(os_info.path.settings, 'colours_domains.json'), 'w') as f:
                        json.dump(domain_colour_dict, f)
                    status_message = 'Saved colours_domains.json'
                if 'words' in arguments:
                    with open(os.path.join(os_info.path.settings, 'colours_words.json'), 'w') as f:
                        json.dump(words_colour_dict, f)
                    status_message = 'Saved colours_words.json'
                if 'commands' in arguments:
                    with open(os.path.join(os_info.path.settings, 'commands.json'), 'w') as f:
                        json.dump(commands, f)
                    status_message = 'Saved commands.json'
                if 'username_to_id' in arguments:
                    with open(os.path.join(os_info.path.settings, 'username_to_id.json'), 'w') as f:
                        json.dump(username_to_id, f)
                    status_message = 'Saved username_to_id.json'
                if 'j' in arguments:
                    with open(os.path.join(os_info.path.settings,
                                           '{0}_{1}_{2}.json'.format(str(j_meta['forum_id']), str(j_meta['number']),
                                                                     j_meta['sorting'])), 'w') as f:
                        json.dump(j, f)
                    status_message = 'Saved j to .json'
                if 'j_forum' in arguments:
                    with open(os.path.join(os_info.path.settings,
                                           '{0}_{1}_{2}.json'.format(str(j_meta['forum_id']), str(j_meta['number']),
                                                                     j_meta['sorting'])), 'w') as f:
                        json.dump(j_forum, f)
                    status_message = 'Saved j_forum to .json'
        elif command in ['url', 'browser'] + settings['browsers']:  # Open forum/thread in web browser
            if command == 'browser':
                command = settings['browsers'][0]  # Use default browser
            command_old = command
            command = arguments[0]
            argument = arguments[1]
            url = ''
            if command[:2] == 'f=':
                url = 'https://www.thestudentroom.co.uk/forumdisplay.php?' + command
            elif command[:2] in ['t=', 'p=']:
                url = 'https://www.thestudentroom.co.uk/showthread.php?' + command
            elif (command[:2] == 'u=') | (command == 'user'):
                try:
                    int(command[2:])
                    url = 'https://www.thestudentroom.co.uk/member.php?' + command
                except:
                    try:
                        url = 'https://www.thestudentroom.co.uk/member.php?' + str(int(argument))
                    except:
                        url = 'https://www.thestudentroom.co.uk/member.php?' + username_to_id[argument]
            elif general.is_number(command):
                thread_id = j[int(command)]['id']
                url = 'https://www.thestudentroom.co.uk/showthread.php?t=' + thread_id
            if command_old == 'url':
                status_message = (url)
            elif url != '':
                subprocess.call([command, url])
        elif command == 'browse':
            if len(arguments) == 0:
                open_pager('''
                Error: Requires at least one argument.

                Usage:
                    browse f [forumid] [page]?
                    browse t [threadid] [page]? [sorting]?
                    browse [threadNumber from forum-like object] [page]?

                Sorting methods include:  reps, date'

                Eg:
                    browse f 143                for page 1 of forum of id 143
                    browse t 12345678 3         for page 3 of thread of id 12345678
                    browse 73 1                 for page 1 of thread of index 143 (index=order in our forum object)

                Tip: Each thread page holds 20 posts
                If thread, also accepts page value of "all"''')
            else:
                sorting = 'datetime'
                if len(arguments) == 0:
                    add_line_('error', 'Requires more arguments')
                    browse_type = 'failed'
                else:
                    if general.is_number(arguments[0]):
                        browse_type = 'thread_list_index'
                        thread_list_index = int(arguments[0])
                        arguments = arguments[1:]
                        if len(arguments) == 0:
                            page_n = settings['default_thread_page']
                        else:
                            if arguments[0] in ['all', 'last']:
                                page_n = arguments[0]
                                arguments = arguments[1:]
                            else:
                                try:
                                    page_n = int(arguments[0])
                                    arguments = arguments[1:]
                                except:
                                    page_n = settings['default_thread_page']
                            try:
                                sorting = arguments[0]
                            except:
                                pass
                    elif len(arguments) == 1:
                        add_line_('error', 'Requires more arguments')
                        browse_type = 'failed'
                    else:
                        browsing['browsing'] = True
                        browse_type = arguments[0]
                        browse_id = int(arguments[1])
                        arguments = arguments[2:]
                        if len(arguments) == 0:
                            page_n = settings['default_thread_page']
                        else:
                            if arguments[0] in ['all', 'last']:
                                page_n = arguments[0]
                                arguments = arguments[1:]
                            elif len(arguments) == 1:
                                page_n = int(arguments[0])
                            else:
                                sorting = arguments[0]
                                page_n = int(arguments[1])
                if page_n == 'first':  # Because of settings['default_thread_page']
                    page_n = 1
                if browse_type == 'f':
                    forum_category = [key for key in forums if str(browse_id) in forums[key]][0]
                    persistant_bars['title']['content'] = '{}: {}'.format(forum_category,
                                                                          forums[forum_category][str(browse_id)][
                                                                              'title'])
                    persistant_bars['subtitle']['content'] = 'Page {}'.format(str(page_n))
                    last_object_displayed = 'forum'
                    browsing['forum_id'] = browse_id
                    (thread_meta_list, _, securitytoken) = soups.load_forumpage_threadmeta(browsing['forum_id'],
                                                                                           page_n)  # We shouldnt attempt to load all forum pages
                    height_of_current_displayed_obj, scroll_lines, thread_meta_list = display_forum(thread_meta_list,
                                                                                                    sorting=sorting)
                elif browse_type in ['t', 'thread_list_index']:
                    if browse_type == 't':
                        title = '?'
                        page_n = 1
                        thread_id = int(browse_id)
                    else:
                        thread_id = int(thread_meta_list[thread_list_index][
                                            'id'])  # Requires us to have run "browse f $FORUM_ID" before!
                        try:
                            title = thread_meta_list[thread_list_index]['title']
                        except:
                            title = thread_meta_list[thread_list_index]['posts'][0]['title']
                        if general.is_number(page_n):
                            page_n = min(page_n, ceil(thread_meta_list[thread_list_index]['nPosts_claimed'] / 20))
                        if page_n == 'last':
                            page_n = ceil(thread_meta_list[thread_list_index][
                                              'nPosts_claimed'] / 20)  # total number of pages in thread
                    persistant_bars['subtitle']['content'] = '{}: page {}'.format(title, str(page_n))
                    last_object_displayed = 'thread'
                    browsing['thread_id'] = thread_id
                    if page_n == 'all':
                        page_n = ceil(thread_meta_list[thread_list_index][
                                          'nPosts_claimed'] / 20)  # total number of pages in thread
                        if browsing['logged_in']:
                            (posts_list, securitytoken) = soups.load_thread(thread_id, page_n, session=session)
                        else:
                            (posts_list, securitytoken) = soups.load_thread(thread_id, page_n)
                    else:
                        if browsing['logged_in']:
                            (posts_list, securitytoken) = soups.load_thread_page(thread_id, page_n, session=session)
                        else:
                            (posts_list, securitytoken) = soups.load_thread_page(thread_id, page_n)
                    height_of_current_displayed_obj, scroll_lines, posts_list = display_posts(posts_list, sorting)
                    username_to_id_new = {post['user']['name']: post['user']['id'] for post in posts_list}
                    username_to_id = dict(username_to_id, **username_to_id_new)  # Combine dicts
                    del username_to_id_new

                    browsing['thread_id'] = thread_id
                    del thread_id

                    if j_meta['forum_id'] != browsing['forum_id']:
                        j = []
                    j = [post for post in j if
                         post['title'] != posts_list[0]['title']]  # So we can replace it with the updated thread
                    j.append({
                        'datetime': posts_list[0]['datetime'],
                        'user': {
                            'name': posts_list[0]['user']['name'],
                            'id': posts_list[0]['user']['id']
                        },
                        'title': posts_list[0]['title'],
                        'nPosts': len(posts_list),
                        'nPosts_claimed': len(posts_list),
                        'posts': posts_list
                    })
            page_n = browsing['page_n']
        elif command == 'login':
            session, account_credentials = login(securitytoken)
        elif command == 'logout':
            logout_url = soups.get_logout_url_from_session(
                session)  # includes the logouthash and everything else required to log out
            session.get(logout_url)
        elif command == 'post':
            status_message = post_comment(session, thread_id=browsing['thread_id'])
        elif command == 'notifications':
            if not browsing['logged_in']:
                session, account_credentials = login(securitytoken)
                browsing['logged_in'] = True
            notifications_and_hrefs = soups.notifications_from_session(session)
            toprint = ''
            height_of_current_displayed_obj = display_tuple_ls(notifications_and_hrefs)
        elif command == 'report':  # Not working yet
            if len(arguments) < 2:
                open_pager('''
                Error:  Requires two arguments

                Descr:  Report a post that breaks site rules

                Usage:
                    report p [post_id]
                    report t [thread_id]')

                Eg:
                    report p 12345678   Report a REPLY with post_id 12345678
                    report t 23456789   Report a THREAD with thread_id of 23456789''')
            else:
                if not browsing['logged_in']:
                    session, account_credentials = login(securitytoken)
                    browsing['logged_in'] = True
                if securitytoken == 'guest':
                    (_, securitytoken) = soups.load_thread_page(browsing['thread_id'], 1, session=session,
                                                                data='report_categories')
                categories = {
                    'postinghelp': {
                        'threadmove': 'Thread in wrong forum',
                        'bump': 'Necroposting',
                        'swearing': 'Swear filter avoidance',
                        'foreignlang': 'Foreign language',
                        'txtspk': 'Textspeak',
                        'anon': 'Misuse of anonymous account'
                    },
                    'offensive': {
                        'hatred': 'Racial or religous hatespeech',
                        'discrimination': 'Discrimination',
                        'bullying': 'Bullying',
                        'harsh': 'Harsh language',
                        'insults': 'Insults',
                        'threats': 'Threats'
                    },
                    'nonconstructive': {
                        'nonconstructive': 'Not constructive'
                    },
                    'adspam': {
                        'advertising': 'Advertising',
                        'officialrep': 'Misuse of Official Rep account',
                        'spam': 'Spam'
                    },
                    'privacy': {
                        'privacy': 'Breach of privacy'
                    },
                    'graphic': {
                        'adult': 'Adult content',
                        'violence': 'Promoting violence'
                    },
                    'illegal': {
                        'copyright': 'Copyright infringement',
                        'drugs': 'Drug use',
                        'otherillegal': 'Other illegal activity'
                    },
                    'safety': {
                        'selfharm': 'Referencing self-harm',
                        'suicide': 'Suicide',
                    },
                    'offtopic': {
                        'offtopic': 'Off-topic comment'
                    },
                    'examdetails': {
                        'examdetails': 'Sharing exam/interview details'
                    }
                }
                data = {
                    'securitytoken': securitytoken,
                    'do': 'doadd',
                    'type': 'post',
                    'primaryid': arguments[0],
                    'message': arguments[1]
                }
                categories_strs = []
                toprint = ''
                for entry in categories:
                    for key in categories[entry]:
                        categories_strs.append(key)
                        toprint += general.make_n_char_long(key, 30) + categories[entry][key]
                open_pager(toprint)
                category = ''
                cancel_action = True
                while category not in categories_strs + ['cancel']:
                    category = input('Category:    ')
                    if category == 'cancel':
                        cancel_action = False
                        break
                if not cancel_action:
                    for entry in categories:
                        for key in categories[entry]:
                            if key == category:
                                if key == entry:
                                    data['category'] = [key]
                                else:
                                    data['category'] = [entry, key]
                    url = 'https://www.thestudentroom.co.uk/report.php'
                    session.post(url, data)
        elif command == 'rep':
            if len(arguments) == 0:
                open_pager('''
                Error:  Requires at least one argument

                Descr:  Award rep to a helpful post

                Usage:  rep [post_id]''')
            elif len(arguments) == 1:
                if not browsing['logged_in']:
                    session, account_credentials = login(securitytoken)
                    browsing['logged_in'] = True
                if securitytoken == 'guest':
                    (_, securitytoken) = soups.load_thread_page(browsing['thread_id'], 1, session=session)
                data = {
                    'securitytoken': securitytoken,
                    'do': 'helpfulanswer',
                    'using_ajax': '1',
                    'rank': '1',
                    'p': arguments[0]
                }
                url = 'https://www.thestudentroom.co.uk/helpfulanswers.php'
                session.post(url, data)
        elif command == 'register':  # Not implemented yet (requires CSS for human verification)
            if len(arguments) != 2:
                open_pager('''
                Error:  Requires at least two arguments

                Descr:  Register a new account

                Usage:  register [username] [password]''')
            elif len(arguments) == 2:
                data = {
                    'securitytoken': securitytoken,
                    'do': 'addmember',
                    'url': 'forum.php',
                    'gpu': '0',
                    'direct': '0',
                    'survey': '0',
                    'username': prompt_input('Username:    '),
                    'password': prompt_input('Password:    '),
                    'year': prompt_input('BIRTH: Year:    '),
                    'month': prompt_input('BIRTH: Month:    '),
                    'day': prompt_input('BIRTH: Day:    '),
                    'agree': '1',
                    'emailoptin': '1',  # 1 seems to mean 'opt out'.
                    'humanverify': {
                        'hash': '',
                        'input': ''
                    }
                }
                session = session_post('register.php', data, logout=True, login=False)
        elif command == 'account':  # Not working yet
            if arguments[0] == 'info':
                if len(arguments) == 0:
                    open_pager('''
                    Requires at least one argument

                    Rep: Award rep to a helpful post

                    Usage:
                        rep [post_id]
                    ''')
                elif len(arguments) == 1:
                    if not browsing['logged_in']:
                        session, account_credentials = login(securitytoken)
                        browsing['logged_in'] = True
                    list_of_flags, securitytoken = soups.get_list_of_flags_and_security_token(session)
                    data = {
                        'securitytoken': securitytoken,
                        'do': 'updateprofile'
                    }
                    data['showbirthday'] = input('Show birthday? 1=yes, 0=no:    ')
                    data['customtext'] = input('Custom user title:    ')
                    data['homepage'] = input('Your website/blog/etc homepage (if any):    ')
                    data['userfield'] = {}
                    data['userfield']['field1'] = input('About me:    ')
                    data['userfield']['field2'] = input('Location of residence:    ')
                    data['userfield']['field8'] = input('Sex (1:M, 2:F, 3:None):    ')
                    if data['userfield']['field8'] == '3':
                        data['userfield']['sexprivacy'] = '0'
                    else:
                        data['userfield']['sexprivacy'] = '1'
                    data['userfield']['field10'] = input('Location of study:    ')
                    data['userfield']['field11'] = input('Academic info:    ')
                    data['userfield']['field12'] = input('Name:    ')
                    data['userfield']['field16'] = input('Custom section:  Title:    ')
                    data['userfield']['field17'] = input('Custom section: Content:    ')
                    data['userfield']['field21'] = input('Sexual orientation:    ')
                    # Now to choose a flag, from too many options

                    open_pager('\n'.join(['  '.join(l) for l in list_of_flags]))  # Display scrolling list of flags
                    data['userfield']['field9'] = input('Flag ID:    ')

                    data['userfield']['field1_set'] = '1'
                    data['userfield']['field2_set'] = '1'
                    data['userfield']['field8_set'] = '1'
                    data['userfield']['field9_set'] = '1'
                    data['userfield']['field10_set'] = '1'
                    data['userfield']['field11_set'] = '1'
                    data['userfield']['field12_set'] = '1'
                    data['userfield']['field16_set'] = '1'
                    data['userfield']['field17_set'] = '1'
                    data['userfield']['field21_set'] = '1'

                    url = 'https://www.thestudentroom.co.uk/profile.php?do=updateprofile'
                    request = session.post(url, data)
                    os_info.writeif(os_info.path.log, request.text, settings['readonly'], append=True)
                    del request
        elif command in ['edit', 'delete']:
            if len(arguments) < 1:
                open_pager('''
                Requires at least one argument')

                edit [post_id] [new_content] [reason]?
                delete [post_id] [reason]?''')
            else:
                if not browsing['logged_in']:
                    session, account_credentials = login(securitytoken)
                    browsing['logged_in'] = True
                post_id = arguments[0]
                if command == 'edit':
                    message = arguments[1]
                    arguments = [arguments[0], arguments[1]]
                if len(arguments) > 2:
                    reason = arguments[2]
                else:
                    reason = ''
                securitytoken, posthash = soups.get_securitytoken_and_posthash_from_url(session,
                                                                                        'https://www.thestudentroom.co.uk/editpost.php?p={0}&do=editpost'.format(
                                                                                            post_id))
                if command == 'edit':
                    data = {
                        'securitytoken': securitytoken,
                        'do': 'updatepost',
                        'emailupdate': '0',
                        'parseurl': '1',
                        'reason': reason,
                        'p': post_id,
                        'posthash': posthash,
                        'message': message
                    }
                    url = 'https://www.thestudentroom.co.uk/editpost.php?do=updatepost'
                elif command == 'delete':
                    data = {
                        'securitytoken': securitytoken,
                        'do': 'deletepost',
                        'reason': reason,
                        'p': post_id,
                        'posthash': posthash,
                        'deletepost': 'delete'
                    }
                    url = 'https://www.thestudentroom.co.uk/editpost.php?do=deletepost'
                session.post(url, data)
        elif command == 'load':  # in commands['forumdisplay']['commands']['forum']['alts']: # Not elif, so that previous forum can be returned to easily
            if len(arguments) == 0:
                open_pager('''
                Error:  Requires at least one argument')

                Usage:  load [forum/f/user/u] [forumid/username] [options]?

                Eg:
                        load f 143')
                        load u MrUsername
                        load f 63 --force       (to reload even if it is already loaded)''')
            else:
                browsing[
                    'browsing'] = False  # So we can load more meta information when displaying the forum-like object
                command = arguments[0]
                arguments = arguments[1:]
                failed = True
                if command in ['f', 'forum']:
                    last_object_displayed = 'forum'
                    forum_type = 'forum'
                    try:
                        forum_id = str(int(arguments[0]))
                        failed = False
                    except:
                        list_forum_ids()
                elif command in ['u', 'user', 'username']:
                    last_object_displayed = 'forum'
                    forum_type = 'userid'
                    failed = False
                    try:
                        forum_id = str(username_to_id[arguments[0]])
                    except:
                        try:
                            forum_id = str(int(arguments[0]))
                        except:
                            open_pager('''
                    Error:  No userid found for username

                    Try loading the forum, or browsing a thread, in which the user was active, and then executing, to expand the list of userids''')
                            failed = True
                if not failed:
                    persistant_bars['title']['content'] = '{}:    {}'.format(command, arguments[0])
                    persistant_bars['subtitle']['content'] = 'Cache'
                    try:
                        number_of_threads = int(arguments[1])
                        try:
                            sorting = arguments[2]
                        except:
                            sorting = 'datetime'
                    except:
                        number_of_threads = 500
                        try:
                            sorting = arguments[1]
                        except:
                            sorting = 'datetime'
                    if (forum_id, number_of_threads, sorting) != (
                    j_meta['forum_id'], j_meta['number'], j_meta['sorting']):
                        if ('--force' in arguments) | (
                                (forum_id, number_of_threads) != (j_meta['forum_id'], j_meta['number'])):
                            (_, _, _, _, j_forum, username_to_id_new) = load_forum_database(forum_type, forum_id,
                                                                                            max_threads=number_of_threads,
                                                                                            sorting=sorting,
                                                                                            load_threads=True)
                            username_to_id = dict(username_to_id, **username_to_id_new)  # Combine dicts
                            del username_to_id_new
                            j_meta['forum_id'] = forum_id
                            j_meta['number'] = number_of_threads
                        j_meta['sorting'] = sorting
                    j = j_forum
                    height_of_current_displayed_obj, scroll_lines, j = display_forum(j, sorting=sorting)
        elif command == 'filter':
            if len(arguments) < 2:
                open_pager('''
                Error:  Requires at least two arguments')

                Descr:  Discard all posts in the forum which do not match the filter

                Usage:  filter [postAttribute/"user"+userattribute] [attributeValue]

                Eg:     filter user MrUsername''')
            else:
                filter_attr = arguments[0]
                attr_is_user_attr = False
                if filter_attr[:4] == 'user':
                    filter_attr = filter_attr[4:]
                    attr_is_user_attr = True
                filter_value = arguments[1]
                try:
                    sorting = arguments[2]
                except:
                    sorting = 'datetime'
                j = []
                for thread in j_forum:
                    thread_new = {x: thread[x] for x in thread if x != 'posts'}
                    thread_new['posts'] = []
                    for post in thread['posts']:
                        if attr_is_user_attr:
                            if filter_attr in post['user']:
                                if post['user'][filter_attr] == filter_value:
                                    thread_new['posts'].append(post)
                        else:
                            if filter_attr in post:
                                if post[filter_attr] == filter_value:
                                    thread_new['posts'].append(post)
                    if len(thread_new['posts']) > 0:
                        j.append(thread_new)
                height_of_current_displayed_obj, scroll_lines, j = display_forum(j, sorting=sorting)
        elif command == 't':
            if len(arguments) == 0:
                open_pager('''
                Error:  Requires at least one argument

                Usgae:  t [thread_id]''')
            else:
                thread_id = int(arguments[0])
                posts_list = [thread for thread in j if j[thread]['id'] == thread_id][0]
                height_of_current_displayed_obj, scroll_lines, posts_list = display_posts(dummy, arguments)
        elif command in ['byday', 'bydate']:
            if len(arguments) == 0:
                open_pager('''
                Error:  Requires at least one argument

                Descr:  Display statistics about a certain post attribute.

                Usage:  byday [postAttribute]

                postAttributes include: word noun title user quote domain

                Eg:
                    byday username
                    byday word
                    byday noun''')
            else:
                words_to_ignore = regex.words_to_ignore + ['png', 'jpg', 'jpeg', 'gif', 'webm', 'mp4']
                try:
                    attr = arguments[0]
                    if attr[:4] == 'user':
                        attr_is_user_attr = True
                        attr = attr[4:]
                    else:
                        attr_is_user_attr = False
                except:
                    attr = 'word'
                dates = []
                for thread in j:
                    for post in thread['posts']:
                        dates.append(general.standardise_date(
                            post['datetime']))  # not sure why it has been failing on one forum and not the other
                dates = set(dates)  # rm duplicates
                dates = sorted(dates, reverse=True)  # Most recent first
                words_counters = {}
                words_counter_monthly = Counter([])
                toprint_monthly = []
                # toprint_tuple_ls=[]
                toprint = ''
                current_line_y = 0
                mainpad.clear()

                if attr in ['word', 'title', 'noun', 'user', 'quote',
                            'quoted']:  # Untagged usernames share the word dictionary with words.
                    colour_dict = dict(
                        words_colour_dict)  # Dont want changes to be saved in words_colour_dict, which might be saved to disk
                elif attr in ['domain']:
                    colour_dict = dict(domain_colour_dict)

                scroll_lines = [0]
                for i in range(len(dates)):
                    date = dates[i]
                    words_counter = Counter([])
                    current_line_y += 2
                    scroll_lines.append(current_line_y)
                    mainpad.addstr(current_line_y, 0, str(date)[:10], str2colour['date'])
                    current_line_x = 11
                    for thread in j:
                        for post in thread['posts']:
                            if date == general.standardise_date(post[
                                                            'datetime']):  # Do NOT use "is" instead of "==". Will only return first post for each date.
                                if attr_is_user_attr:
                                    words = [post['user'][attr]]
                                else:
                                    content = '\n'.join([x['content'] for x in post['contents'] if x[
                                        'type'] != 'quote'])  # We don't want to muddle the results by including quotes
                                    if attr == 'word':
                                        words = regex.ls_words(content)
                                    elif attr == 'noun':
                                        if use_nltk:
                                            content = ' '.join(re.split(regex.between_words, content))
                                            words_nltk_tokenised = nltk.word_tokenize(content)
                                            words_nltk_tagged = nltk.pos_tag(words_nltk_tokenised)

                                            words_nltk_tagged = [(w[0].lower(), w[1]) for w in words_nltk_tagged]

                                            nouns = [w[0] for w in words_nltk_tagged if w[1] == 'NN']
                                            verbs = [w[0] for w in words_nltk_tagged if w[1] in ['VBG', 'VB']]
                                            misc = [w[0] for w in words_nltk_tagged if w[1] == 'JJ']
                                            stemmer = nltk.PorterStemmer()
                                            nouns = [stemmer.stem(w) for w in nouns]
                                            verbs = [stemmer.stem(w) for w in verbs]
                                            misc = [stemmer.stem(w) for w in misc]
                                            words = nouns + verbs + misc
                                            words = [w for w in words if (len(w) > 2) & (w not in words_to_ignore)]
                                        else:
                                            open_pager(
                                                'Error:   nltk modules nltk.word_tokenize, nltk.pos_tag, nltk.PorterStemmer not installed')
                                            words = []
                                    elif attr == 'title':
                                        words = re.split(regex.between_words_title, post['thread']['title'])
                                        words = [w.lower() for w in words]
                                        words = [singularise(w) for w in words if
                                                 (len(w) > 2) & (w not in words_to_ignore)]
                                    elif attr == 'domain':
                                        words = regex.findall_domains(content)
                                    elif attr in ['quote', 'quotes']:
                                        words = [x['user'] for x in post['quotes']]
                                words_counter = words_counter + Counter(words)
                    words = []
                    words_countls = [(words_counter[x], x) for x in
                                     words_counter]  # list of tuples (username, #occurances)
                    results = sorted(words_countls, key=lambda x: x[0], reverse=True)[:10]
                    # result=' '.join(results)
                    n_colour = 0
                    for tpl in results:
                        word = tpl[1]
                        freq = tpl[0]
                        if word not in words_colour_dict:
                            if 'only-tagged' not in arguments:
                                colour_dict[word] = cl_to_assign[n_colour % len(cl_to_assign)]
                                n_colour += 1
                            else:
                                colour_dict[word] = 'default'
                        mainpad.addstr(current_line_y, current_line_x, word, str2colour[colour_dict[word]])
                        mainpad.addstr(current_line_y + 1, current_line_x, str(freq), str2colour[colour_dict[word]])
                        current_line_x += len(word) + 1
                    words_counter_monthly = words_counter_monthly + words_counter
                    words_counter = Counter([])

                    print_monthly_counter = False
                    if i + 1 == len(dates):
                        print_monthly_counter = True
                    elif date.month != dates[i + 1].month:
                        print_monthly_counter = True
                    if print_monthly_counter:
                        words_countls = [(words_counter_monthly[x], x) for x in
                                         words_counter_monthly]  # list of tuples (username, #occurances)
                        results = sorted(words_countls, key=lambda x: x[0], reverse=True)[:10]
                        n_colour = 0
                        toprint_monthly.append((str(date.year) + '-' + str(date.month), results))
                        words_counter_monthly = Counter([])
                for tupl in toprint_monthly:  # So we get monthly results after the daily results
                    current_line_y += 2
                    scroll_lines.append(current_line_y)
                    mainpad.addstr(current_line_y, 0, tupl[0], str2colour['date'])
                    current_line_x = 10
                    for tpl in tupl[1]:
                        word = tpl[1]
                        freq = tpl[0]
                        if word not in colour_dict:
                            if 'only-tagged' not in arguments:
                                colour_dict[word] = cl_to_assign[n_colour % len(cl_to_assign)]
                                n_colour += 1
                            else:
                                colour_dict[word] = 'default'
                        mainpad.addstr(current_line_y, current_line_x, word, str2colour[colour_dict[word]])
                        mainpad.addstr(current_line_y + 1, current_line_x, str(freq), str2colour[colour_dict[word]])
                        current_line_x += len(word) + 1
        elif command in ['grep', 'egrep']:
            if len(arguments) == 0:
                open_pager('''
                Error: Requires at least one argument.

                Usage: [grep/egrep] [pattern]

                eg: grep exam
                grep "biology exam"
                egrep "(aqa|AQA) [bB]iology"''')
            else:
                posts_list = []
                search_string = arguments[0]
                if command == 'egrep':
                    regexp = re.compile(search_string)
                for thread in j:
                    for post_old in thread['posts']:
                        post = dict(post_old)
                        print_post = False
                        for content_type in post['contents']:
                            post_content = content_type['content']
                            post_contents = post_content.split('\n')
                            for line in post_contents:
                                if len(line) > 0:
                                    if line[0] != '>':
                                        if command == 'grep':
                                            if search_string in line:
                                                print_post = True
                                                content_type['content'] = content_type['content'].replace(search_string,
                                                                                                          '@@@grep~~~~~~' + search_string + '@@@default~~~')
                                        if command == 'egrep':
                                            if regexp.search(line):
                                                print_post = True
                                                content_type['content'] = re.sub(r'({0})'.format(search_string),
                                                                                 r'{0}\1{1}'.format('@@@grep~~~~~~',
                                                                                                    '@@@default~~~'),
                                                                                 content_type['content'],
                                                                                 flags=re.MULTILINE)  # For this line, I must acknowledge Martijn Pieters for his helpful answer on StackOverflow.
                        if print_post:
                            posts_list.append(post)
                height_of_current_displayed_obj, scroll_lines, posts_list = display_posts(posts_list)
        elif command == 'ifrep':  # eg "reps >10" to show all posts with rep > 10, "reps" to show all posts sorted by rep
            if len(arguments) < 2:
                open_pager('''
                Error:  Requires at least two arguments

                Usage:  ifrep [equivalence relation] [number]

                Eg:
                        ifrep > 30
                        ifrep = 10''')
            else:
                equivalence = arguments[0]
                n = int(arguments[1])
                posts_list = []
                for thread in j:
                    for post in thread['posts']:
                        print_post = False
                        if equivalence == '>':
                            if post['reps'] > n:
                                print_post = True
                        elif equivalence == '<':
                            if post['reps'] < n:
                                print_post = True
                        elif equivalence in ['=', '==']:
                            if post['reps'] == n:
                                print_post = True
                        if print_post:
                            posts_list.append(post)
                height_of_current_displayed_obj, scroll_lines, posts_list = display_posts(posts_list)
        elif command == 'debug':
            if arguments[0] == 'path':
                open_pager('paths = ' + str(os_info.paths))  # The dictionary of paths that the path class is built from
            elif arguments[0] == 'dict':
                if arguments[1] == 'users':
                    open_pager(str(users))
        elif command == 'sort':  # Sort current displayed object by an attribute
            if len(arguments) == 1:
                sorting = arguments[0]
                if last_object_displayed == 'thread':
                    height_of_current_displayed_obj, scroll_lines, posts_list = display_posts(posts_list, sorting)
                elif last_object_displayed == 'forum':
                    if browsing['browsing']:
                        height_of_current_displayed_obj, scroll_lines, thread_meta_list = display_forum(
                            thread_meta_list, sorting=sorting)
                    else:
                        height_of_current_displayed_obj, scroll_lines, j = display_forum(j, sorting=sorting)
            else:
                open_pager('''
                Error:  Requires one argument

                Usage:  sort [attribute]

                Attributes include:
                        username
                        userid
                        userreps
                        reps    (reps of post)
                        datetime''')
        else:
            if general.is_number(command):
                n = int(command)
                try:
                    sorting = arguments[0]
                except:
                    sorting = 'datetime'
                attr_is_user_attr = False
                if sorting[:4] == 'user':
                    attr_is_user_attr = True
                    sorting = sorting[4:]
                posts_list = j[n]['posts']
                if attr_is_user_attr:
                    posts_list = sorted(posts_list, key=lambda x: x['user'][sorting],
                                        reverse=True)  # Highest at the top
                else:
                    posts_list = sorted(posts_list, key=lambda x: x[sorting], reverse=True)
                height_of_current_displayed_obj, scroll_lines, posts_list = display_posts(posts_list)
                last_object_displayed = 'thread'
            else:
                pass


def main():
    curses.wrapper(mainbody)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import json, io, re, os, sys, shutil, argparse, requests, time
from bs4 import BeautifulSoup

sys.path.insert(0, os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
from ..modules.general import *
from ..modules.os_info import *
from ..modules import regex, forumload, soups

parser = argparse.ArgumentParser()
parser.add_argument('--port', type=int)
parser.add_argument('--forum-id', type=int)
parser.add_argument('--user-id', type=int)
parser.add_argument('--thread_type', type=str)
parser.add_argument('--page-first', type=int)
parser.add_argument('--page-last', type=int)
parser.add_argument('--overwrite-threads', dest='overwrite_threads', action='store_true')
parser.add_argument('--no-overwrite-threads', dest='overwrite_threads', action='store_false')
parser.add_argument('--overwrite-posts', dest='overwrite_posts', action='store_true')
parser.add_argument('--no-overwrite-posts', dest='overwrite_posts', action='store_false')
parser.add_argument('--delay', type=float)

mkdir(path.main)  # Superfluous, since mkdir() is recursive, but ignoring would look odd
mkdir(path.posts)
mkdir(path.threads)
mkdir(path.users)
mkdir(path.forums)
mkdir(path.settings)
touch(path.log)

settings = json.load(open(os.path.join(path.settings, 'settings.json')))


class Port_Source(requests.adapters.HTTPAdapter):
    # Credit: Martijn Pieters https://stackoverflow.com/questions/47202790/python-requests-how-to-specify-port-for-outgoing-traffic
    def __init__(self, port, *args, **kwargs):
        self._source_port = port
        super(Port_Source, self).__init__(*args, **kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = requests.packages.urllib3.poolmanager.PoolManager(
            num_pools=connections, maxsize=maxsize,
            block=block, source_address=('', self._source_port))


def main_raw(**kwargs):
    try:
        threads = json.load(
            open(path.threads_json))  # json of all threads ever cached with json caching setting enabled
    except:
        threads = {}
    previously_cached_post_ids = []
    for t in threads:
        print(t)
        if 'posts' in t:
            for p in t['posts']:
                previously_cached_post_ids.append(p)

    def parse_threadpage(url):
        print('parse_threadpage({})'.format(url))
        soup = BeautifulSoup(session.get(url).text, 'html.parser')
        if '?t=' in url:
            thread_id = int(url.split('?t=')[-1].split('&')[0])
        else:
            thread_id = int(soup.find('link', {'rel': 'canonical'})['href'].split('?t=')[-1].split('&')[0])
        print(
            'soups.parse_threadpage({}, {}, {}, overwrite_posts={})'.format('soup', url, thread_type, overwrite_posts))
        open('/tmp/' + url.split('/')[-1] + '.0.html', 'w').write(str(soup))
        (post_list, _) = soups.parse_threadpage(soup, url, thread_type, overwrite_posts=overwrite_posts)
        del soup
        if len(post_list) == 20:
            page_last = page_current + 1  # tmp, lazy workaround
            print(len(post_list))
        else:
            page_last = 1
        return page_last

    def parse_threadpages(thread_type, thread_or_post_id):
        page_current = 1
        page_last = 1
        if thread_type == 'p':
            url = 'https://www.thestudentroom.co.uk/showthread.php?p={}'.format(thread_or_post_id)
            try:
                thread_id = soups.get_thread_id(BeautifulSoup(session.get(url).text, 'html.parser'))
            except:
                thread_id = None  # Thread probably removed
            time.sleep(delay)
        if thread_type == 't':
            thread_id = thread_or_post_id
        if thread_id:
            while page_current <= page_last:
                url = 'https://www.thestudentroom.co.uk/showthread.php?t={}&page={}'.format(thread_id, page_current)
                page_last = parse_threadpage(url)
                print(
                    'Parsed {} {} - Page {}/{}'.format(thread_type, str(thread_id), str(page_current), str(page_last)))
                page_current += 1
                time.sleep(delay)

    # Defaults/init states
    forum_id = False
    user_id = False
    thread_type = 't'
    page_first = 1
    page_last = 999999
    overwrite_threads = True
    overwrite_posts = False
    delay = 11.0

    if kwargs['forum_id'] != None:
        forum_id = kwargs['forum_id']
    elif kwargs['user_id'] != None:
        user_id = kwargs['user_id']
    if kwargs['thread_type'] != None:
        thread_type = kwargs['thread_type']
    if kwargs['page_first'] != None:
        page_first = int(kwargs['page_first'])
    if kwargs['page_last'] != None:
        page_last = int(kwargs['page_last'])
    if kwargs['overwrite_threads'] != None:
        overwrite_threads = kwargs['overwrite_threads']
    if kwargs['overwrite_posts'] != None:
        overwrite_posts = kwargs['overwrite_posts']
    if kwargs['delay'] != None:
        delay = int(kwargs['delay'])
    session = requests.Session()
    if 'port' in kwargs:
        if kwargs['port'] != None:
            session.mount('https://', Port_Source(kwargs['port']))
    d = {}
    thread_ids = []
    threads_cannot_load = []
    page_current = page_first
    if forum_id:
        id_value = forum_id
        filter_type = 'forumid'
    elif user_id:
        id_value = user_id
        filter_type = 'userid'
    else:
        raise Exception('Please specify either a forum-id or user-id')
    while page_current <= page_last:
        print('Requesting search page {}'.format(str(page_current)))
        if thread_type == 't':
            url = 'https://www.thestudentroom.co.uk/search.php?filter[type]=thread&filter[{}]={}&page={}'.format(
                str(filter_type), str(id_value), str(page_current))
        elif thread_type == 'p':
            url = 'https://www.thestudentroom.co.uk/search.php?filter[{}]={}&page={}'.format(str(filter_type),
                                                                                             str(id_value),
                                                                                             str(page_current))
        request = session.get(url)
        soup = BeautifulSoup(request.text, 'html.parser')
        thread_meta_list = soups.load_searchpage_threadmeta(soup, thread_type=thread_type)
        for thread in thread_meta_list:
            if thread_type == 't':
                thread_id = thread['id']
                if thread_id in threads:
                    nPosts_actual = len(threads[thread_id]['posts'])
                else:
                    nPosts_actual = -9  # Might have nPosts_claimed==0
                if nPosts_actual < thread['nPosts_claimed']:
                    print('Parsing t {}'.format(str(thread_id)))
                    parse_threadpages('t', thread_id)
            elif thread_type == 'p':
                post_id = thread['postid']
                if post_id in previously_cached_post_ids:
                    nPosts_actual = 999  # No way of knowing nPosts_claimed
                else:
                    nPosts_actual = -9  # Might have nPosts_claimed==0
                if nPosts_actual < thread['nPosts_claimed']:
                    print('Parsing thread which contains p {}'.format(str(post_id)))
                    parse_threadpages('p', post_id)
            time.sleep(delay)
        page_current += 1
        page_last = int(soup.find('a', {'class': 'pager-last'})['href'].split('&page=')[-1].split('&')[0])
        print('Last page {}'.format(str(page_last)))


def main():
    args = sys.argv[1:]
    main_raw(**vars(parser.parse_args(args)))  # So we can treat it as a dictionary


if __name__ == '__main__':
    main()



import datetime, json
from .os_info import path
import os.path
from dateutil.relativedelta import relativedelta # Since datetime now refuses to work with months and years

users={}
username_colour_dict={} # temporary colour scheme for users without flairs/colours, for easier reading
username_to_id={}
users=json.load(open(os.path.join(path.settings, 'users.json')))
for entry in users:
    if 'colour' not in users[entry]:
        users[entry]['colour']='end'
settings=json.load(open(os.path.join(path.settings, 'settings.json')))



def level_domain(level, domain):
    if not is_int(level):
        raise ValueError('level_domain(level_number, domain_string)')
    neglevel=-int(level)
    return '.'.join(domain.split('.')[neglevel:])



def dictfind(original_dict, **kwargs):
    ''' eg
    >>>d={
        'foo':{
            'bar': 11,
            'bob': 'george',
        },
        'gimli':{
            'gandalf':{
                'frodo':5,
                'samwise':7,
            },
            'radagast':11,
        },
        'jones':{
            'gandalf':{
                'frodo':1,
                'samwise':2,
            },
            'radagast':3,
        },
    }
    >>>find(d, bar='>5')
    ['foo']
    >>>find(d, gandalf_frodo='>3')
    ['gimli']
    '''
    # NB foo='*bar*' will not work
    ListOnlyKeys=True
    if '_ListOnlyKeys' in kwargs:
        ListOnlyKeys=kwargs['_ListOnlyKeys']
    d=dict(original_dict) # Apparently dicts are automatically global
    for kw in kwargs:
        if kw[0]!='_':
            kw_value=kwargs[kw]
            kw_relation=None
            if kw_value[0] in ['*', '>', '<', '=']:
                kw_relation=kw_value[0]
                kw_value=kw_value[1:]
            kw_relation_end=None
            if kw_value[-1] in ['*']:
                kw_relation_end=kw_value[-1]
                kw_value=kw_value[:-1]
            new_d={}
            key0=None
            if '_' in kw:
                key0=kw.split('_')[0]
                key1=kw.split('_')[1]
                for key in d:
                    if key0 in d[key]:
                        '''for key in d[key0]:
                            if kw_relation=='*':
                                if key.startswith(kw_value):
                                    new_d[key]=d[key]
                            elif kw_relation_end=='*':
                                if key.endswith(kw_value):
                                    new_d[key]=d[key]'''
                        if key1 in d[key][key0]:
                            this_value=d[key][key0][key1]
                            if kw_relation=='>':
                                kw_value=int(kw_value)
                                if this_value>kw_value:
                                    new_d[key]=d[key]
                            elif kw_relation=='<':
                                kw_value=int(kw_value)
                                if this_value>kw_value:
                                    new_d[key]=d[key]
                            elif kw_relation=='=':
                                if this_value==kw_value:
                                    new_d[key]=d[key]
                '''Following elif segment is more efficient (since not check each loop), but worse code'''
            elif kw_relation_end=='*':
                kw_value=kw_value[:-1]
                for key in d:
                    if key.startswith(kw_value):
                        new_d[key]=d[key]
                    #if key.startswith(kw_value):
                    #    ls.append( key )
            elif kw_relation=='*':
                kw_value=kw_value
                for key in d:
                    if key.endswith(kw_value):
                        new_d[key]=d[key]
                        #ls.append( key )
            elif kw_relation=='>':
                kw_value=int(kw_value)
                for key in d:
                    if d[key][kw]>kw_value:
                        print('       {}>{}'.format(d[key][kw], kw_value), end='\r')
                        new_d[key]=d[key]
                        #ls.append( key )
            elif kw_relation=='<':
                kw_value=int(kw_value)
                for key in d:
                    if d[key][kw]<kw_value:
                        new_d[key]=d[key]
                        #ls.append( key )
            else:
                for key in d:
                    if d[key][kw]==kw_value:
                        new_d[key]=d[key]
            
            d=dict(new_d)
    if ListOnlyKeys:
        return [key for key in d]
    else:
        return d


def userid_colour(userid):
    if type(userid)!='str':
        userid=str(userid)
    try:
        return users[userid]['colour']
    except:
        raise Exception('userid {0} has no colour. users[{0}] = {1}'.format(userid, users[userid]))


def make_n_char_long(x, n, spacing=' '):
    y=str(x)
    while len(y)<n:
        y+=spacing
    if len(y)>n:
        y=y[:n]
    return y


def myjoin(*args):
    string=''
    for arg in args:
        string+='  '+str(arg)
    return string[2:]


def standardise_datetime(datestr):
    if isinstance(datestr, datetime.datetime):
        return datestr
    elif isinstance(datestr, datetime.date):
        return datetime.datetime.combine(datestr, datetime.time(0, 0)) # Sets to midnight
    try:
        return datetime.datetime.strptime(datestr, '%d-%m-%Y')
    except:
        try:
            return datetime.datetime.strptime(datestr, '%Y-%m-%d')
        except:
            datestr=datestr.strip().lower()
            if datestr=='yesterday':
                return datetime.datetime.now()-datetime.timedelta(days=1)
            elif 'ago' in datestr:
                date_ls=datestr.split(' ')
                #print(date_ls)
                try:
                    n=int(date_ls[0])
                    formatIs_n_obj_ago=True
                except:
                    formatIs_n_obj_ago=False
                if formatIs_n_obj_ago:
                    date_type=date_ls[1]
                    if date_type in ['second', 'seconds']:
                        return datetime.datetime.now()-datetime.timedelta(seconds=n)
                    elif date_type in ['minute', 'minutes']:
                        return datetime.datetime.now()-relativedelta(minutes=n)
                    elif date_type in ['hour', 'hours']:
                        return datetime.datetime.now()-relativedelta(hours=n)
                    elif date_type in ['day', 'days']:
                        return datetime.datetime.now()-datetime.timedelta(days=n)
                    elif date_type in ['week', 'weeks']:
                        return datetime.datetime.now()-datetime.timedelta(days=n*7)
                    elif date_type in ['month', 'months']:
                        return datetime.datetime.now()-relativedelta(months=n)
                    elif date_type in ['year', 'years']:
                        return datetime.datetime.now()-relativedelta(years=n)
                    elif date_type in ['decade', 'decades']:
                        return datetime.datetime.now()-relativedelta(years=n*10)
            else:
                for char in ['T', ' ', '_']:
                    if len(datestr)>18: # So probably includes seconds
                        try:
                            try:
                                datestr=datestr[:19] # Remove the trailing +00:00
                                return datetime.datetime.strptime(datestr, '%Y-%m-%d'+char+'%H:%M:%S')
                            except:
                                datestr=datestr[:19] # Remove the trailing +00:00
                                return datetime.datetime.strptime(datestr, '%d-%m-%Y'+char+'%H:%M:%S')
                        except:
                            pass
                    else:
                        try:
                            try:
                                return datetime.datetime.strptime(datestr, '%Y-%m-%d'+char+'%H:%M')
                            except:
                                return datetime.datetime.strptime(datestr, '%d-%m-%Y'+char+'%H:%M')
                        except:
                            pass
    raise Exception('Unknown datetime string: '+str(datestr))


def standardise_datetime_str(datestr):
    return str(standardise_datetime(datestr))


def standardise_date(string):
    return standardise_datetime(string).date()


def standardise_date_str(datestr):
    return str(standardise_date(datestr))


def standardise_time(timestr):
    return datetime.datetime.strptime(timestr, '%H:%M').time()


def is_number(x):
    try:
        dummy=float(x)
        return True
    except:
        return False

def is_int(x):
    try:
        dummy=int(x)
        return True
    except:
        return False

def ls_rm_dupl(ls): # Useful to have rather than list(set(, since it preserves order. Perhaps sorted(list(set( would be more performative, I will look into this eventually.
    l=[]
    for x in ls:
        if x not in l:
            l.append(x)
    return l


def ls_in_str(ls,string):
    for element in ls:
        try:
            if element in string:
                return True
        except:
            pass
    return False


def mystrip(text):
    while '  ' in text:
        text=text.replace('  ',' ')
    while '\n\n' in text:
        text=text.replace('\n\n','\n')
    while ' \n' in text:
        text=text.replace(' \n','\n')
    while '\n ' in text:
        text=text.replace('\n ','\n')
    while '\n> > ' in text:
        text=text.replace('\n> > ','\n> ')
    while '>\n' in text:
        text=text.replace('>\n','')
    while '~\n' in text:
        text=text.replace('~\n','\n')
    while '\n\n' in text:
        text=text.replace('\n\n','\n')
    while '\r\n' in text:
        text=text.replace('\r\n','\n')
    while '\n\r' in text:
        text=text.replace('\n\r','\n')
    while '\n\n' in text:
        text=text.replace('\n\n','\n')
    return text.strip()



try:
    import inflect
    use_inflect=True
except ImportError:
    use_inflect=False

if use_inflect:
    inflect_engine=inflect.engine()
def singular_noun(word):
    if use_inflect:
        try:
            return inflect_engine.singular_noun(word)
        except:
            pass
    return word
def singular_verb(word):
    if use_inflect:
        try:
            return inflect_engine.singular_verb(word)
        except:
            pass
    return word

def singularise(word):
    x=singular_noun(word) # Placeholder for if/when I find a package for this
    if type(x).__name__=='str':
        return x
    else:
        return word


import re
from . import grammar
from .os_info import path
from .general import make_n_char_long
import os.path
import json # for json.load


def regexise_forename(forename):
    return '['+forename[0].lower()+forename[0].upper()+']'+forename[1:]


def build_colour_regexes():
    colours_regex=json.load(open(os.path.join(path.settings, 'colours_regex.json')))
    colours_and_regex=[]
    for entry in colours_regex:
        d=colours_regex[entry]
        if 'regex' in d:
            regex='(^|[ \(\)\[\]]*)('+d['regex']+')('+between_words+')'
            groups=[1, 2, re.compile(regex).groups]
            colours_and_regex.append((d['colour'], regex, groups))
        if 'forenames_optional' in d:
            names_regex=[]
            groups_n=1
            for name in d['forenames_optional']:
                namesplit=name.split(' ')
                if len(namesplit)==0:
                    pass
                elif len(namesplit)==1:
                    regex='({0})'.format(regexise_forename(name))
                    groups_n+=1
                    names_regex.append(regex)
                else:
                    forename=name.split(' ')[0]
                    nameparts=[regexise_forename(namepart) for namepart in name.split(' ')[1:]]
                    nameparts=' '.join(nameparts)
                    regex='(({0} )?({1}))'.format(regexise_forename(forename), nameparts)
                    groups_n+=3
                    names_regex.append(regex)
            regex='(^|[ \(\)\[\]]*)((([mM]rs?)|([pP]resident))? ?'+'|'.join(names_regex)+')('+between_words+')'
            groups=[1, 2, groups_n+5] # group number of: start characters, name, end characters
            colours_and_regex.append((d['colour'], regex, groups))
        if 'allnames' in d:
            names_regex=[]
            groups_n=1
            for name in d['allnames']:
                namesplit=name.split(' ')
                if len(namesplit)==1:
                    regex='({0})'.format(regexise_forename(name))
                    names_regex.append(regex)
                    groups_n+=1
                else:
                    namesplit=name.split(' ')
                    forename=name.split(' ')[0]
                    nameparts=[regexise_forename(namepart) for namepart in name.split(' ')[1:]]
                    nameparts=' '.join(nameparts)
                    regex='({0} {1})'.format(regexise_forename(forename), nameparts)
                    groups_n+=1
                    names_regex.append(regex)
            regex='(^|[ \(\)\[\]]*)('+'|'.join(names_regex)+')('+between_words+')'
            groups=[1, 2, groups_n+2] # group number of: start characters, name, end characters
            colours_and_regex.append((d['colour'], regex, groups))
    return colours_and_regex


def parse_custom_colours(string):
    for tpl in colours_and_regex:
        colour_string='@@@'+make_n_char_long(tpl[0], 10, '~')
        regex=tpl[1]
        groups=tpl[2]
        string=re.sub(regex, r'\{0}{1}\{2}{3}\{4}'.format(groups[0], colour_string, groups[1], '@@@useTwoLeft', groups[2]), string, flags=re.MULTILINE)
    return string

url='https?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'
url2=r'(https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{2,256}\.[a-z]{2,6}\b([-a-zA-Z0-9@:%_\+.,~#?&//=]*))'
image=r'([a-zA-Z0-9_-]*\.(jpg|jpeg|png|gif))'
between_words=r'[\ \n\.\,\;\:\/\-\+\=\?\!\"\&\(\)\[\]\@\£\$\%\’]'
between_words_title=r'[\ \n\.\,\;\:\/\-\+\=\?\!\"\&\(\)\[\]\@\£\$\%]'

# For future use analysing style of texts
sentence_ending=' ?[.\!\?$]*' # Some users use ..., some ?!!, some nothing

words_to_ignore="we're maybe aren't you'd goes where it's wouldn't can't many him didn't isn't i'd i've won't lol tho he'll imo too etc wouldn't you're him these mean far such http https dont thing also got com www use youre now say even into the and that have with was you for their more her not back had some are really she his this one men who doesn't think think most people but they would don't like just has all should it's all just what our can will how then your then only get when how from he's i'm any did much because its other them there than yes may been want does about said why those been could very going which being see know were well out we've out that's she's".split(' ')


def domain_from_url(url):
    url=re.sub(r'(https?://((www|images?|img|[a-z])\.)?)', r'', url)
    url=url.split('/')[0]
    try:
        return re.search(r'([a-z][a-z]*\.(com|(co|org|gov|police|ac)\.(uk|nz|au)|org|gov|net|tech|ac|edu|io|ie|de|fr|ca|cz|ru|se|gr|nz|nl|gl|eu|tv|va|pl))', url).group(0)
    except:
        return url


def findall_domains(search_string, **kwargs):
    if 'only_image_hosts' not in kwargs:
        kwargs['only_image_hosts']=False
    domains=[]
    urls = re.findall(url, search_string)
    for u in urls:
        if kwargs['only_image_hosts']:
            add_domain=False
            if u.split('.')[-1].lower() in ['jpg', 'png', 'jpeg', 'gif', 'webm', 'mp4']:
                add_domain=True
        else:
            add_domain=True
        if add_domain:
            domains.append(domain_from_url(u))
    return domains

def findall_urls(search_string):
    return re.findall(url, search_string)

def ls_words(string):
    words=re.split(between_words, string)
    words=[w.lower() for w in words]
    words_new=[]
    for w in words:
        if words[0]+words[-1] in ["''", '""', '``']:
            words_new.append(w[1:-1])
        else:
            words_new.append(w)
    words=words_new
    words=[grammar.singularise(w) for w in words if (len(w)>2)&(w not in words_to_ignore)]
    return words


colours_and_regex=build_colour_regexes()
